import itertools
import os
from distutils.spawn import find_executable

import mdtraj as md
import numpy as np
import pytest
from mdtraj.geometry.dihedral import (compute_chi1, compute_chi2, compute_chi3,
                                      compute_chi4, compute_omega, compute_phi,
                                      compute_psi, compute_chi5)
from mdtraj.testing import eq


@pytest.fixture()
def traj(get_fn):
    return md.load(get_fn('2EQQ.pdb'))


@pytest.fixture()
def pymol():
    pymol = find_executable('pymol')
    if pymol is None:
        try_paths = ['/Applications/MacPyMOL.app/Contents/MacOS/MacPyMOL']
        for path in try_paths:
            if os.path.exists(path):
                pymol = path
                continue

    if pymol is None:
        raise pytest.skip("pymol executable not found")
    return pymol


def test_compute_phi(traj):
    rx, x = compute_phi(traj, opt=True)
    ry, y = compute_phi(traj, opt=False)
    eq(rx, ry)
    eq(x, y)


def test_compute_psi(traj):
    rx, x = compute_psi(traj, opt=True)
    ry, y = compute_psi(traj, opt=False)
    eq(rx, ry)
    eq(x, y)


def test_compute_chi(get_fn, tmpdir, pymol, traj):
    """Test that the proper chi angles are computed.

    The "reference data" is manually computed with pymol.
    """
    try:
        from jinja2 import Template
    except ImportError:
        pytest.skip("jinja2 is required for this test")

    pymol_script = Template('''
import numpy as np
from pymol import cmd, CmdException

CHI1_ATOMS = [["N", "CA", "CB", "CG"],
              ["N", "CA", "CB", "CG1"],
              ["N", "CA", "CB", "SG"],
              ["N", "CA", "CB", "OG"],
              ["N", "CA", "CB", "OG1"]]
CHI2_ATOMS = [["CA", "CB", "CG", "CD"],
            ["CA", "CB", "CG", "CD1"],
            ["CA", "CB", "CG1", "CD1"],
            ["CA", "CB", "CG", "OD1"],
            ["CA", "CB", "CG", "ND1"],
            ["CA", "CB", "CG", "SD"]]
CHI3_ATOMS = [["CB", "CG", "CD", "NE"],
            ["CB", "CG", "CD", "CE"],
            ["CB", "CG", "CD", "OE1"],
            ["CB", "CG", "SD", "CE"]]
CHI4_ATOMS = [["CG", "CD", "NE", "CZ"],
            ["CG", "CD", "CE", "NZ"]]

CHI5_ATOMS = ["CD", "NE", "CZ", "NH1"]

for n, chi_atoms in enumerate([CHI1_ATOMS, CHI2_ATOMS, CHI3_ATOMS, CHI4_ATOMS, CHI5_ATOMS]):
    indices, angles = [], []
    for r in range(1, cmd.count_atoms('name ca')+1):
        r_indices, r_angles = None, None
        for atoms in chi_atoms:
            try:
                selection = ['%d/%s' % (r, e) for e in atoms]
                r_indices = [cmd.id_atom(s) - 1 for s in selection]
                r_angles =  cmd.get_dihedral(*selection) * np.pi / 180.0
            except CmdException:
                pass
        if r_indices is not None:
            indices.append(r_indices)
            angles.append(r_angles)
    np.savetxt("{{indices_fn}}.%d" % (n+1), indices, '%d')
    np.savetxt("{{angles_fn}}.%d" % (n+1), angles)

cmd.quit()''')

    indices_fn = os.path.join(tmpdir, 'indices')
    angles_fn = os.path.join(tmpdir, 'angles')
    pymolscript_fn = os.path.join(tmpdir, 'pymolscript.py')

    with open(pymolscript_fn, 'w') as f:
        f.write(pymol_script.render({'indices_fn': indices_fn,
                                     'angles_fn': angles_fn}))

    os.system('%s %s -cr %s' % (pymol, get_fn('2EQQ.pdb'), pymolscript_fn))

    ref_indices = [np.loadtxt(indices_fn + '.%d' % i, dtype=int) for i in range(1, 5)]
    ref_angles = [np.loadtxt(angles_fn + '.%d' % i) for i in range(1, 5)]

    indices, angles = zip(*[compute_chi1(traj, opt=False), compute_chi2(traj, opt=False),
                            compute_chi3(traj, opt=False), compute_chi4(traj, opt=False),
                            compute_chi5(traj, opt=False)])
    indices_opt, angles_opt = zip(*[compute_chi1(traj, opt=True), compute_chi2(traj, opt=True),
                                    compute_chi3(traj, opt=True), compute_chi4(traj, opt=True),
                                    compute_chi5(traj, opt=True)])

    for x, y in zip(indices, ref_indices):
        eq(x, y)
    for x, y in zip(indices_opt, ref_indices):
        eq(x, y)
    for x, y in zip(angles, ref_angles):
        eq(x[0], y, decimal=4)
    for x, y in zip(angles_opt, ref_angles):
        eq(x[0], y, decimal=4)


def test_compute_omega(traj):
    rx, x = compute_omega(traj, opt=True)
    ry, y = compute_omega(traj, opt=False)
    eq(rx, ry)
    eq(x, y)


def test_shape_when_none(get_fn):
    t = md.load(get_fn('frame0.h5'))
    np.hstack((md.compute_phi(t)[1],
               md.compute_psi(t)[1],
               md.compute_chi1(t)[1],
               md.compute_chi2(t)[1],
               md.compute_chi3(t)[1],
               md.compute_chi1(t)[1],
               md.compute_omega(t)[1]))


def test_dihedral_1(pymol, tmpdir):
    xyz = '''MODEL        0
ATOM      1    A ACE     1       4.300  13.100   8.600  1.00  0.00
ATOM      2    B ACE     1       5.200  13.600   8.800  1.00  0.00
ATOM      3    C ACE     1       4.900  14.300   9.600  1.00  0.00
ATOM      4    D ACE     1       5.600  14.200   7.900  1.00  0.00
    '''
    script = '''
from pymol import cmd
with open('output.txt', 'w') as f:
    f.write('%f' % cmd.get_dihedral('1/A', '1/B', '1/C', '1/D'))
'''
    prevdir = os.path.abspath('.')
    try:
        os.chdir(tmpdir)
        with open('xyz.pdb', 'w') as f:
            f.write(xyz)
        with open('pymolscript.py', 'w') as f:
            f.write(script)

        os.system('%s %s -cr %s' % (pymol, 'xyz.pdb', 'pymolscript.py'))
        with open('output.txt') as f:
            pymol_value = np.deg2rad(float(f.read()))
        t = md.load('xyz.pdb')
    finally:
        os.chdir(prevdir)

    mdtraj_value = md.compute_dihedrals(t, [[0, 1, 2, 3]])[0, 0]

    np.testing.assert_array_almost_equal(pymol_value, mdtraj_value)


def test_dihedral_2chains(get_fn):
    # make sure that comput_phi is finding dihedrals from all of the chains
    # in a multi-chain topology
    t = md.load_pdb(get_fn('4OH9.pdb'))

    # remove the water
    water_indices = [a.index for a in t.top.atoms if a.residue.name != 'HOH']
    t.restrict_atoms(water_indices)

    # okay we've got two protein chains
    assert t.top.n_chains == 2
    phi_indices, angles = md.compute_phi(t)

    for chain in t.top.chains:
        chain_indices = [a.index for a in chain.atoms]

        # assert that at least one of the phi_indices involves atoms in this chain
        assert any(i in chain_indices for i in np.concatenate(phi_indices))


def test_generator():
    N_FRAMES = 2
    N_ATOMS = 5
    xyz = np.asarray(np.random.randn(N_FRAMES, N_ATOMS, 3), dtype=np.float32)
    ptraj = md.Trajectory(xyz=xyz, topology=None)

    quartets = np.array(list(itertools.combinations(range(N_ATOMS), 4)), dtype=np.int32)
    quartets2 = itertools.combinations(range(N_ATOMS), 4)
    a = md.compute_dihedrals(ptraj, quartets)
    b = md.compute_dihedrals(ptraj, quartets2)
    eq(a, b)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Tim Moore
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np

from mdtraj.testing import eq
from mdtraj import load


def test_load(get_fn):
    t = load(get_fn('test_hoomdxml.dcd'), top=get_fn('well-mixed.hoomdxml'))
    top = load(get_fn('well-mixed.hoomdxml')).topology
    eq(t.topology, top)
    eq(top.n_atoms, 6600)
    eq(top.n_bonds, 3200)
    eq(top.n_chains, 4000)
    eq(top.n_chains, top.n_residues)  # hoomdxml makes each chain 1 residue


def test_load_no_chains(get_fn):
    top = load(get_fn('no_chains.hoomdxml')).topology
    eq(top.n_atoms, 3)
    eq(top.n_bonds, 0)
    eq(top.n_chains, top.n_atoms)
    eq(top.n_chains, top.n_residues)

def test_load_no_ions(get_fn):
    top = load(get_fn('no_ions.hoomdxml')).topology
    eq(top.n_atoms, 1296)
    eq(top.n_bonds, 1152)
    eq(top.n_chains, 144)
    eq(top.n_chains, top.n_residues)
    for bond in top.bonds:  # atoms bonded to adjacent atoms by index
        eq(bond[1].index - bond[0].index, 1)

def test_correct_ordering(get_fn):
    traj = load(get_fn('water-box.hoomdxml'))
    sodium = traj.top.select("name 'Na'")
    sodium_slice = traj.atom_slice(sodium)
    assert np.isclose(sodium_slice.xyz, np.array([[[30, 30, 30]]])).all()



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: John D. Chodera and Kyle A. Beauchamp, Kim Branson,
#          Imran Haque, Michael Shirts
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# This code is adapted from Yank: GPU-accelerated calculation of ligand
# binding affinities in implicit and explicit solvent using alchemical
# free energy methodologies, copyright 2009-2011 University of California,
# Berkeley, Vertex Pharmaceuticals, Stanford University, University of Virginia,
# and the Authors. Yank is licensed under the GPU Lesser General Public
# License.
##############################################################################

import numpy as np
from mdtraj.utils import check_random_state

__all__ = ["rotation_matrix_from_quaternion", "uniform_quaternion"]



def rotation_matrix_from_quaternion(q):
    """Compute a 3x3 rotation matrix from a given quaternion (4-vector).
    
    Parameters
    ----------
    q : np.ndarray, size=(..., 4)
        Quaternion or array of quaternions (need not be normalized, zero norm OK)

    Returns
    -------
    Rq : np.ndarray, size=(..., 3, 3)
        Orthogonal rotation matrices corresponding to quaternion(s) q. Given,
        for example, an input shape of (m, 4), Rq will have shape (m, 3, 3).

    Examples
    --------
    >>> q = np.array([0.1, 0.2, 0.3, -0.4])
    >>> Rq = rotation_matrix_from_quaternion(q)

    See Also
    --------
    uniform_quaternion

    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Rotation_matrix#Quaternion
    """

    w = q[..., 0]; x = q[..., 1]; y = q[..., 2]; z = q[..., 3]
    # w, x, y, z = q
    Nq = np.sum(q**2, axis=q.ndim-1)

    if q.shape == (4,):
        s = 2.0 / Nq if Nq > 0 else 0.0
    else:
        with np.errstate(divide='ignore'):
            s = 2.0 / Nq
            s[Nq <= 0.0] = 0.0

    X = x*s; Y = y*s; Z = z*s
    wX = w*X; wY = w*Y; wZ = w*Z
    xX = x*X; xY = x*Y; xZ = x*Z
    yY = y*Y; yZ = y*Z; zZ = z*Z
    Rq = np.array([[ 1.0-(yY+zZ),       xY-wZ,        xZ+wY  ],
                   [      xY+wZ,   1.0-(xX+zZ),       yZ-wX  ],
                   [      xZ-wY,        yZ+wX,   1.0-(xX+yY) ]])

    if q.shape == (4,):
        return Rq
    return np.rollaxis(Rq, Rq.ndim-1)


def uniform_quaternion(size=None, random_state=None):
    """Generate uniform normalized quaternion 4-vectors

    Parameters
    ----------
    size : int or tuple of ints, optional
        Defines the shape of the returned array of quaternions. If None
        (the default), returns a quaternion 4-vector.
   random_state : integer or numpy.RandomState, optional
        The generator used for random numbers. If an integer is given,
        it fixes the seed. Defaults to the global numpy random number
        generator.

    Returns
    -------
    out : ndarray
        Array of quaternion 4-vectors. Given a ``size`` of, for example,
        ``(m,n)``, ``m*n`` samples are generated.  Because each sample is
        `4-dimensional, the output shape is ``(m,n,4)``. If no shape is
        specified, a single (`4`-D) sample is returned.

    See Also
    --------
    rotation_matrix_from_quaternion

    References
    ----------
    .. [1] K. Shoemake. Uniform random rotations. In D. Kirk, editor, Graphics
       Gems III, pages 124-132. Academic, New York, 1992.
    .. [2] http://planning.cs.uiuc.edu/node198.html
    """
    random = check_random_state(random_state)
    if size is None:
       randsize = (3, 1)
    elif isinstance(size, tuple):
        size = tuple(int(i) for i in size)
        randsize = (3,) + size
    else:
        randsize = (3, int(size))

    u = random.random_sample(randsize)
    q = np.array([np.sqrt(1 - u[0]) * np.sin(2*np.pi * u[1]),
                  np.sqrt(1 - u[0]) * np.cos(2*np.pi * u[1]),
                  np.sqrt(u[0])     * np.sin(2*np.pi * u[2]),
                  np.sqrt(u[0])     * np.cos(2*np.pi * u[2])])

    if size is None:
        return q[:, 0]
    return np.rollaxis(q, 0, q.ndim)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors: Robert McGibbon, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import os
import pickle
import tempfile

import mdtraj as md
import numpy as np
import pytest
from mdtraj.testing import eq

try:
    from simtk.openmm import app
    import simtk.unit as u

    HAVE_OPENMM = True
except ImportError:
    HAVE_OPENMM = False

needs_openmm = pytest.mark.skipif(not HAVE_OPENMM, reason='needs OpenMM')


@needs_openmm
def test_topology_openmm(get_fn):
    topology = md.load(get_fn('1bpi.pdb')).topology
    topology_with_bond_order = md.load(get_fn('imatinib.mol2')).topology

    # the openmm trajectory doesn't have the distinction
    # between resSeq and index, so if they're out of whack
    # in the openmm version, that cant be preserved
    for top in [topology, topology_with_bond_order]:
        for residue in top.residues:
            residue.resSeq = residue.index
        mm = top.to_openmm()
        assert isinstance(mm, app.Topology)
        topology2 = md.Topology.from_openmm(mm)
        eq(top, topology2)


@needs_openmm
def test_topology_openmm_boxes(get_fn):
    traj = md.load(get_fn('1vii_sustiva_water.pdb'))
    mmtop = traj.topology.to_openmm(traj=traj)
    box = mmtop.getUnitCellDimensions() / u.nanometer


def test_topology_pandas(get_fn):
    topology = md.load(get_fn('native.pdb')).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)

    # Make sure default argument of None works, see issue #774
    topology3 = md.Topology.from_dataframe(atoms)


def test_topology_pandas_TIP4PEW(get_fn):
    topology = md.load(get_fn('GG-tip4pew.pdb')).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)


def test_topology_pandas_2residues_same_resSeq(get_fn):
    topology = md.load(get_fn('two_residues_same_resnum.gro')).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)


def test_topology_numbers(get_fn):
    topology = md.load(get_fn('1bpi.pdb')).topology
    assert len(list(topology.atoms)) == topology.n_atoms
    assert len(list(topology.residues)) == topology.n_residues
    assert all([topology.atom(i).index == i for i in range(topology.n_atoms)])


def test_topology_unique_elements_bpti(get_fn):
    traj = md.load(get_fn('bpti.pdb'))
    top, bonds = traj.top.to_dataframe()
    atoms = np.unique(["C", "O", "N", "H", "S"])
    eq(atoms, np.unique(top.element.values))


def test_chain(get_fn):
    top = md.load(get_fn('bpti.pdb')).topology
    chain = top.chain(0)
    assert chain.n_residues == len(list(chain.residues))

    atoms = list(chain.atoms)
    assert chain.n_atoms == len(atoms)
    for i in range(chain.n_atoms):
        assert atoms[i] == chain.atom(i)


def test_residue(get_fn):
    top = md.load(get_fn('bpti.pdb')).topology
    residue = top.residue(0)
    assert len(list(residue.atoms)) == residue.n_atoms
    atoms = list(residue.atoms)
    for i in range(residue.n_atoms):
        assert residue.atom(i) == atoms[i]


def test_segment_id(get_fn):
    top = md.load(get_fn('ala_ala_ala.pdb')).topology
    assert next(top.residues).segment_id == "AAL", "Segment id is not being assigned correctly for ala_ala_ala.psf"
    df = top.to_dataframe()[0]
    assert len(df["segmentID"] == "AAL") == len(
        df), "Segment id is not being assigned correctly to topology data frame ala_ala_ala.psf"


def test_nonconsective_resSeq(get_fn):
    t = md.load(get_fn('nonconsecutive_resSeq.pdb'))
    assert eq(np.array([r.resSeq for r in t.top.residues]), np.array([1, 3, 5]))
    df1 = t.top.to_dataframe()
    df2 = md.Topology.from_dataframe(*df1).to_dataframe()
    assert eq(df1[0], df2[0])

    # round-trip through a PDB load/save loop
    fd, fname = tempfile.mkstemp(suffix='.pdb')
    os.close(fd)
    t.save(fname)
    t2 = md.load(fname)
    assert eq(df1[0], t2.top.to_dataframe()[0])
    os.unlink(fname)


def test_pickle(get_fn):
    # test pickling of topology (bug #391)
    topology_without_bond_order = md.load(get_fn('bpti.pdb')).topology
    topology_with_bond_order = md.load(get_fn('imatinib.mol2')).topology
    for top in [topology_with_bond_order, topology_without_bond_order]:
        loaded_top = pickle.loads(pickle.dumps(top))
        assert loaded_top == top


def test_atoms_by_name(get_fn):
    top = md.load(get_fn('bpti.pdb')).topology

    atoms = list(top.atoms)
    for atom1, atom2 in zip(top.atoms_by_name('CA'), top.chain(0).atoms_by_name('CA')):
        assert atom1 == atom2
        assert atom1 in atoms
        assert atom1.name == 'CA'

    assert len(list(top.atoms_by_name('CA'))) == sum(1 for _ in atoms if _.name == 'CA')
    assert top.residue(15).atom('CA') == [a for a in top.residue(15).atoms if a.name == 'CA'][0]

    with pytest.raises(KeyError):
        top.residue(15).atom('sdfsdf')


def test_select_atom_indices(get_fn):
    top = md.load(get_fn('native.pdb')).topology

    assert eq(top.select_atom_indices('alpha'), np.array([8]))
    assert eq(top.select_atom_indices('minimal'),
              np.array([4, 5, 6, 8, 10, 14, 15, 16, 18]))

    with pytest.raises(ValueError):
        top.select_atom_indices('sdfsdf')


@needs_openmm
def test_top_dataframe_openmm_roundtrip(get_fn):
    t = md.load(get_fn('2EQQ.pdb'))
    top, bonds = t.top.to_dataframe()
    t.topology = md.Topology.from_dataframe(top, bonds)
    omm_top = t.top.to_openmm()


def test_n_bonds(get_fn):
    t = md.load(get_fn('2EQQ.pdb'))
    for atom in t.top.atoms:
        if atom.element.symbol == 'H':
            assert atom.n_bonds == 1
        elif atom.element.symbol == 'C':
            assert atom.n_bonds in [3, 4]
        elif atom.element.symbol == 'O':
            assert atom.n_bonds in [1, 2]


def test_load_unknown_topology(get_fn):
    try:
        md.load(get_fn('frame0.dcd'), top=get_fn('frame0.dcd'))
    except IOError as e:
        # we want to make sure there's a nice error message than includes
        # a list of the supported topology formats.
        assert all(s in str(e) for s in ('.pdb', '.psf', '.prmtop'))
    else:
        assert False  # fail


def test_unique_pairs():
    n = 10
    a = np.arange(n)
    b = np.arange(n, n + n)

    eq(md.Topology._unique_pairs(a, a).sort(), md.Topology._unique_pairs_equal(a).sort())
    eq(md.Topology._unique_pairs(a, b).sort(), md.Topology._unique_pairs_mutually_exclusive(a, b).sort())


def test_select_pairs(get_fn):
    traj = md.load(get_fn('tip3p_300K_1ATM.pdb'))
    select_pairs = traj.top.select_pairs

    assert len(select_pairs(selection1='name O', selection2='name O')) == 258 * (258 - 1) // 2
    assert len(select_pairs(selection1='name H1', selection2='name O')) == 258 * 258

    selections = iter([
        # Equal
        ("(name O) or (name =~ 'H.*')", "(name O) or (name =~ 'H.*')"),
        ('all', 'all'),

        # Exclusive
        ('name O', 'name H1'),
        ('name H1', 'name O'),

        # Overlap
        (range(traj.n_atoms), 'name O'),
        ('all', 'name O')])

    for select1, select2 in selections:
        select3, select4 = next(selections)
        assert eq(select_pairs(selection1=select1, selection2=select2).sort(),
                  select_pairs(selection1=select3, selection2=select4).sort())


def test_to_fasta(get_fn):
    t = md.load(get_fn('2EQQ.pdb'))
    assert t.topology.to_fasta(0) == "ENFSGGCVAGYMRTPDGRCKPTFYQLIT"


def test_subset(get_fn):
    t1 = md.load(get_fn('2EQQ.pdb')).top
    t2 = t1.subset([1, 2, 3])
    assert t2.n_residues == 1

def test_subset_re_index_residues(get_fn):
    t1 = md.load(get_fn('2EQQ.pdb')).top
    t2 = t1.subset(t1.select('resid 0 2'))
    np.testing.assert_array_equal([0, 1], [rr.index for rr in t2.residues])


def test_molecules(get_fn):
    top = md.load(get_fn('4OH9.pdb')).topology
    molecules = top.find_molecules()
    assert sum(len(mol) for mol in molecules) == top.n_atoms
    assert sum(1 for mol in molecules if len(mol) > 1) == 2  # All but two molecules are water


def test_copy_and_hash(get_fn):
    t = md.load(get_fn('traj.h5'))
    t1 = t.topology
    t2 = t.topology.copy()
    assert t1 == t2

    assert hash(tuple(t1._chains)) == hash(tuple(t2._chains))
    assert hash(tuple(t1._atoms)) == hash(tuple(t2._atoms))
    assert hash(tuple(t1._bonds)) == hash(tuple(t2._bonds))
    assert hash(tuple(t1._residues)) == hash(tuple(t2._residues))

    assert hash(t1) == hash(t2)


def test_topology_sliced_residue_indices(get_fn):
    # https://github.com/mdtraj/mdtraj/issues/1585
    full = md.load(get_fn('1bpi.pdb'))
    residues = full.top.select("resid 1 to 10")
    sliced = full.atom_slice(residues)
    idx = [res.index for res in sliced.top.residues][-1]
    assert idx == sliced.top.n_residues-1
    # Now see if this works
    _ = sliced.topology.residue(idx)

def test_topology_join(get_fn):
    top_1 = md.load(get_fn('2EQQ.pdb')).topology
    top_2 = md.load(get_fn('4OH9.pdb')).topology

    out_topology = top_1.join(top_2)

    eq(out_topology.n_atoms, top_1.n_atoms + top_2.n_atoms)
    eq(out_topology.n_residues, top_1.n_residues + top_2.n_residues)
    eq(top_1.atom(0).residue.name, out_topology.atom(0).residue.name)
    eq(top_2.atom(-1).residue.name, out_topology.atom(-1).residue.name)
    eq(top_1.atom(0).element, out_topology.atom(0).element)
    eq(top_2.atom(-1).element, out_topology.atom(-1).element)

def test_topology_join_keep_resSeq(get_fn):
    top_1 = md.load(get_fn('2EQQ.pdb')).topology
    top_2 = md.load(get_fn('4OH9.pdb')).topology

    out_topology_keepId_True = top_1.join(top_2, keep_resSeq=True)
    out_topology_keepId_False = top_1.join(top_2, keep_resSeq=False)

    out_resSeq_keepId_True = [residue.resSeq for residue in out_topology_keepId_True.residues]
    out_resSeq_keepId_False = [residue.resSeq for residue in out_topology_keepId_False.residues]

    expected_resSeq_keepId_True = (
        [residue.resSeq for residue in top_1.residues
            ] + [
                residue.resSeq for residue in top_2.residues])
    expected_resSeq_keepId_False = list(range(1, len(expected_resSeq_keepId_True) + 1))

    eq(out_resSeq_keepId_True, expected_resSeq_keepId_True)
    eq(out_resSeq_keepId_False, expected_resSeq_keepId_False)



# cython: c_string_type=str, c_string_encoding=ascii
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


###############################################################################
# Imports
###############################################################################

import cython
import warnings
cimport cython
from libc.stdio cimport SEEK_SET, SEEK_CUR, SEEK_END
import os
import numpy as np
cimport numpy as np
np.import_array()
from mdtraj.utils import ensure_type, cast_indices, in_units_of
from mdtraj.utils.six import string_types
from mdtraj.formats.registry import FormatRegistry
from libc.stdlib cimport malloc, free
from binposlib cimport molfile_timestep_t
from binposlib cimport seek_timestep, tell_timestep;
from binposlib cimport open_binpos_read, close_file_read, read_next_timestep
from binposlib cimport open_binpos_write, close_file_write, write_timestep

###############################################################################
# Globals
###############################################################################

__all__ = ['BINPOSTrajectoryFile', 'load_binpos']
# codes that indicate status on return from library
cdef int _BINPOS_SUCESS = 0  # regular exit code
cdef int _BINPOS_EOF = -1  # end of file (or error)

###############################################################################
# Classes
###############################################################################

@FormatRegistry.register_loader('.binpos')
def load_binpos(filename, top=None, stride=None, atom_indices=None, frame=None):
    """load_binpos(filename, top=None, stride=None, atom_indices=None, frame=None)

    Load an AMBER .binpos file from disk.

    The .binpos format is a cross-platform binary trajectory format produced by
    AMBER software. It stores only the atomic coordinates, and does *not* store
    any unit cell informations. Its use is discouraged.

    Parameters
    ----------
    filename : path-like
        Path of AMBER binpos file.
    top : {path-like, Trajectory, Topology}
        The BINPOS format does not contain topology information. Pass in either
        the path to a pdb file, a trajectory, or a topology to supply this
        information.
    stride : int, default=None
        Only read every stride-th frame
    atom_indices : array_like, optional
        If not none, then read only a subset of the atoms coordinates from the
        file. This may be slightly slower than the standard read because it
        requires an extra copy, but will save memory.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.

    Examples
    --------
    >>> import mdtraj as md                                        # doctest: +SKIP
    >>> traj = md.load_binpos('output.binpos', top='topology.pdb') # doctest: +SKIP
    >>> print traj                                                 # doctest: +SKIP
    <mdtraj.Trajectory with 500 frames, 423 atoms at 0x110740a90>  # doctest: +SKIP

    >>> traj2 = md.load_binpos('output.dcd', stride=2, top='topology.pdb') # doctest: +SKIP
    >>> print traj2                                                      # doctest: +SKIP
    <mdtraj.Trajectory with 250 frames, 423 atoms at 0x11136e410>         # doctest: +SKIP

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    See Also
    --------
    mdtraj.BINPOSTrajectoryFile :  Low level interface to BINPOS files
    """
    from mdtraj.core.trajectory import _parse_topology

    # we make it not required in the signature, but required here. although this
    # is a little wierd, its good because this function is usually called by a
    # dispatch from load(), where top comes from **kwargs. So if its not supplied
    # we want to give the user an informative error message
    if top is None:
        raise ValueError('"top" argument is required for load_binpos')
    if not isinstance(filename, (string_types, os.PathLike)):
        raise TypeError('filename must be of type path-like for load_binpos. '
            'you supplied %s' % type(filename))

    topology = _parse_topology(top)
    atom_indices = cast_indices(atom_indices)

    with BINPOSTrajectoryFile(str(filename)) as f:
        if frame is not None:
            f.seek(frame)
            n_frames = 1
        else:
            n_frames = None

        return f.read_as_traj(topology, n_frames=n_frames, stride=stride,
                              atom_indices=atom_indices)


cdef class BINPOSTrajectoryFile:
    """BINPOSTrajectoryFile(filename, mode='r', force_overwrite=True, **kwargs)

    Interface for reading and writing to an AMBER BINPOS file.
    This is a file-like object, that both reading or writing depending
    on the `mode` flag. It implements the context manager protocol,
    so you can also use it with the python 'with' statement.

    The conventional units in the BINPOS file are angstroms. The format only
    supports storing the cartesian coordinates.

    Parameters
    ----------
    filename : str
        The filename to open. A path to a file on disk.
    mode : {'r', 'w'}
        The mode in which to open the file, either 'r' for read or 'w' for write.
    force_overwrite : bool
        If opened in write mode, and a file by the name of `filename` already
        exists on disk, should we overwrite it?

    Other Parameters
    ----------------
    min_chunk_size : int, default=100
        BINPOS, In read mode, we need to allocate a buffer in which to store the data
        without knowing how many frames are in the file. This parameter is the
        minimum size of the buffer to allocate.
    chunk_size_multiplier : int, default=1.5
        In read mode, we need to allocate a buffer in which to store the data
        without knowing how many frames are in the file. We can *guess* this
        information based on the size of the file on disk, but it's not perfect.
        This parameter inflates the guess by a multiplicative factor.

    Examples
    --------
    >>> # copy data from one file to another
    >>> with BINPOSTrajectoryFile('traj.binpos') as f:
    >>>     xyz = f.read()
    >>> with BINPOSTrajectoryFile('out.binpos') as f:
    >>>    f.write(xyz)

    See Also
    --------
    mdtraj.load_binpos : High-level wrapper that returns a ``md.Trajectory``
    """

    cdef int n_atoms
    cdef int n_frames # numnber of frames in the file, cached
    cdef int approx_n_frames
    cdef int min_chunk_size
    cdef int chunk_size_multiplier
    cdef int is_open
    cdef int force_overwrite
    cdef long int frame_counter
    cdef char* mode
    cdef char* filename
    cdef int write_initialized
    cdef readonly char* distance_unit

    cdef void* fh
    cdef molfile_timestep_t* timestep

    def __cinit__(self, char* filename, char* mode='r', force_overwrite=True, **kwargs):
        """Open an AMBER BINPOS file for reading/writing.
        """
        self.distance_unit = 'angstroms'
        self.is_open = False
        self.frame_counter = 0
        self.n_frames = -1
        self.force_overwrite = force_overwrite

        if str(mode) == 'r':
            self.n_atoms = 0
            if not os.path.exists(filename):
                raise IOError("The file '%s' doesn't exist" % filename)
            self.fh = open_binpos_read(filename, "binpos", &self.n_atoms)
            if self.n_atoms <= 0:
                raise IOError('Malformed BINPOS file. Number of atoms <= 0. '
                              'Are you sure this is a valid AMBER BINPOS file?')

            # binpos stores the 4 bytes per float, 3*n_atoms floats per frame, directly
            # with no compression
            self.approx_n_frames = os.stat(filename).st_size / (self.n_atoms * 4 * 3)

            self.min_chunk_size = max(kwargs.pop('min_chunk_size', 100), 1)
            self.chunk_size_multiplier = max(kwargs.pop('chunk_size_multiplier', 1.1), 0.01)
            self.is_open = True

            if self.fh is NULL:
                raise IOError('There was an error opening the binpos file: %s' % filename)

        elif str(mode) == 'w':
            self.write_initialized = False
        else:
            raise ValueError('mode must be one of "r" or "w". you supplied "%s"' % mode)

        self.timestep = <molfile_timestep_t*> malloc(sizeof(molfile_timestep_t))
        if self.timestep is NULL:
            raise MemoryError('There was an error allocating working space')

        for key in kwargs.keys():
            warnings.warn('kwarg "%s" was not recognized or processed' % key)

        self.filename = filename
        self.mode = mode

    def _initialize_write(self, n_atoms):
        force_overwrite = self.force_overwrite
        filename = self.filename
        if force_overwrite and os.path.exists(filename):
            os.unlink(filename)
        if not force_overwrite and os.path.exists(filename):
            raise IOError('"%s" already exists' % filename)
        self.fh = open_binpos_write(filename, "binpos", n_atoms)
        self.n_atoms = n_atoms
        self.is_open = True

    def read_as_traj(self, topology, n_frames=None, stride=None, atom_indices=None):
        """read_as_traj(topology, n_frames=None, stride=None, atom_indices=None)

        Read a trajectory from a BINPOS file

        Parameters
        ----------
        topology : Topology
            The system topology
        n_frames : int, None
            The number of frames you would like to read from the file.
            If None, all of the remaining frames will be loaded.
        stride : int, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object containing the loaded portion of the file.

        See Also
        --------
        read : Returns the raw data from the file
        """
        from mdtraj.core.trajectory import Trajectory
        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        initial = int(self.frame_counter)
        xyz = self.read(n_frames=n_frames, stride=stride, atom_indices=atom_indices)
        if len(xyz) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        in_units_of(xyz, self.distance_unit, Trajectory._distance_unit, inplace=True)

        if stride is None:
            stride = 1
        time = (stride*np.arange(len(xyz))) + initial

        return Trajectory(xyz=xyz, topology=topology, time=time)

    def read(self, n_frames=None, stride=None, atom_indices=None):
        """read(n_frames=None, stride=None, atom_indices=None)

        Read data from a BINPOS file

        Parameters
        ----------
        n_frames : int, None
            The number of frames you would like to read from the file.
            If None, all of the remaining frames will be loaded.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from
            the file. This may be slightly slower than the standard read
            because it required an extra copy, but will save memory.

        Returns
        -------
        xyz : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=np.float32
            The cartesian coordinates, in angstroms
        """
        if not str(self.mode) == 'r':
            raise ValueError('read() is only available when file is opened in mode="r"')

        if n_frames is not None:
            # if they supply the number of frames they want, that's easy
            if not int(n_frames) == n_frames:
                raise ValueError('n_frames must be an int, you supplied "%s"' % n_frames)
            return self._read(int(n_frames), atom_indices)[::stride]

        else:
            all_xyz = []
            while True:
                # guess the size of the chunk to read, based on how many frames we think are in the file
                # and how many we've currently read
                chunk = max(abs(int((self.approx_n_frames - self.frame_counter) * self.chunk_size_multiplier)),
                            self.min_chunk_size)
                xyz = self._read(chunk, atom_indices)

                if len(xyz) <= 0:
                    break
                all_xyz.append(xyz)

            return np.concatenate(all_xyz)[::stride]


    def _read(self, int n_frames, atom_indices=None):
        cdef int n_atoms_to_read, i

        if atom_indices is None:
            n_atoms_to_read = self.n_atoms
        elif isinstance(atom_indices, slice):
            n_atoms_to_read = len(np.arange(self.n_atoms)[atom_indices])
        else:
            if min(atom_indices) < 0:
                raise ValueError('atom_indices should be zero indexed. you gave an index less than zerp')
            if max(atom_indices) >= self.n_atoms:
                raise ValueError('atom indices should be zero indexed. you gave an index bigger than the number of atoms')
            n_atoms_to_read = len(atom_indices)

        # only used if atom_indices is given
        cdef np.ndarray[dtype=np.float32_t, ndim=2] framebuffer = np.zeros((self.n_atoms, 3), dtype=np.float32)

        cdef np.ndarray[dtype=np.float32_t, ndim=3] xyz = np.zeros((n_frames, n_atoms_to_read, 3), dtype=np.float32)

        for i in range(n_frames):
            if atom_indices is None:
                self.timestep.coords = &xyz[i,0,0]
                status = read_next_timestep(self.fh, self.n_atoms, self.timestep)
            else:
                self.timestep.coords = &framebuffer[0, 0]
                status = read_next_timestep(self.fh, self.n_atoms, self.timestep)
                xyz[i, :, :] = framebuffer[atom_indices, :]

            self.frame_counter += 1

            if status != _BINPOS_SUCESS:
                break

        if status != _BINPOS_SUCESS:
            return xyz[:i]
        return xyz

    def write(self, xyz):
        """write(xyz)

        Write cartesian coordinates to a binpos file

        Parameters
        ----------
        xyz : np.ndarray, dtype=np.float32, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of the atoms in every frame, in angstroms.
        """
        xyz = ensure_type(xyz, dtype=np.float32, ndim=3, name='xyz', can_be_none=False,
                          add_newaxis_on_deficient_ndim=True)

        if not self.write_initialized:
            self._initialize_write(xyz.shape[1])
            self.write_initialized = True
        else:
            if not self.n_atoms == xyz.shape[1]:
                raise ValueError('number of atoms in file (%d) does not match number '
                                 'of atoms youre trying to save (%d)' % (self.n_atoms, xyz.shape[1]))

        self._write(xyz)

    def _write(self, np.ndarray[dtype=np.float32_t, ndim=3] xyz not None):
        cdef int i, status
        cdef int n_frames = len(xyz)

        for i in range(n_frames):
            self.timestep.coords = &xyz[i, 0, 0]
            status = write_timestep(self.fh, self.timestep)

            if status != _BINPOS_SUCESS:
                raise RuntimeError("BINPOS Error: %s" % status)

    def seek(self, int offset, int whence=0):
        """seek(offset, whence=0)

        Move to a new file position

        Parameters
        ----------
        offset : int
            A number of frames.
        whence : {0, 1, 2}
            0: offset from start of file, offset should be >=0.
            1: move relative to the current position, positive or negative
            2: move relative to the end of file, offset should be <= 0.
            Seeking beyond the end of a file is not supported
        """
        cdef int origin
        if str(self.mode) != 'r':
            raise NotImplementedError("seek is only supported in mode='r'")

        if whence == 0:
            origin = SEEK_SET
        elif whence == 1:
            origin = SEEK_CUR
        elif whence == 2:
            origin = SEEK_END
        else:
            raise IOError('Invalid argument')

        seek_timestep(self.fh, offset, origin)
        self.frame_counter = tell_timestep(self.fh)

    def tell(self):
        """Current file position

        Returns
        -------
        offset : int
            The current frame in the file.
        """
        return tell_timestep(self.fh)

    def close(self):
        """Close the BINPOS file"""
        if self.is_open:
            if str(self.mode) == 'r':
                close_file_read(self.fh)
            else:
                close_file_write(self.fh)
            self.is_open = False

    def __dealloc__(self):
        free(self.timestep)
        self.close()

    def __enter__(self):
        "Support the context manager protocol"
        return self

    def __exit__(self, *exc_info):
        "Support the context manager protocol"
        self.close()

    def __len__(self):
        if str(self.mode) != 'r':
            raise NotImplementedError('len() only available in mode="r" currently')
        if not self.is_open:
            raise ValueError('I/O operation on closed file')
        if self.n_frames == -1:
            position = self.tell()
            self.seek(0, 2)
            self.n_frames = self.tell()
            self.seek(position)
        return self.n_frames
FormatRegistry.register_fileobject('.binpos')(BINPOSTrajectoryFile)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors

# Authors: Christoph Klein
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import tempfile, os
import numpy as np

import mdtraj as md
from mdtraj.formats import XYZTrajectoryFile
from mdtraj.testing import eq

fd, temp = tempfile.mkstemp(suffix='.xyz')
fd_gz, temp_gz = tempfile.mkstemp(suffix='.xyz.gz')


def teardown_module(module):
    """Remove the temporary file created by tests in this file
    this gets automatically called by pytest. """
    os.close(fd)
    os.unlink(temp)


def test_read_0(get_fn):
    with XYZTrajectoryFile(get_fn('frame0.xyz')) as f:
        xyz = f.read()
    with XYZTrajectoryFile(get_fn('frame0.xyz')) as f:
        xyz3 = f.read(stride=3)
    eq(xyz[::3], xyz3)


def test_read_1(get_fn):
    reference = md.load(get_fn('frame0.dcd'), top=get_fn('native.pdb'))
    traj = md.load(get_fn('frame0.xyz'), top=get_fn('native.pdb'))

    eq(reference.xyz[0], traj.xyz[0], decimal=3)


def test_read_gz(get_fn):
    reference = md.load(get_fn('frame0.dcd'), top=get_fn('native.pdb'))
    traj = md.load(get_fn('frame0.xyz.gz'), top=get_fn('native.pdb'))

    eq(reference.xyz[0], traj.xyz[0], decimal=3)


def test_read_write():
    xyz = np.around(10 * np.random.randn(100, 11, 3), decimals=3)
    with XYZTrajectoryFile(temp, mode='w') as f:
        f.write(xyz)
    with XYZTrajectoryFile(temp) as f:
        xyz2 = f.read()

    eq(xyz, xyz2)


def test_mdwrite(get_fn):
    t = md.load(get_fn('frame0.xyz'), top=get_fn('native.pdb'))
    t.save(temp)
    t.save(temp_gz)


def test_multiread(get_fn):
    reference = md.load(get_fn('frame0.xyz'), top=get_fn('native.pdb'))
    with XYZTrajectoryFile(get_fn('frame0.xyz')) as f:
        xyz0 = f.read(n_frames=1)
        xyz1 = f.read(n_frames=1)

    eq(reference.xyz[0], xyz0[0] / 10)
    eq(reference.xyz[1], xyz1[0] / 10)


def test_seek(get_fn):
    reference = md.load(get_fn('frame0.xyz'), top=get_fn('native.pdb'))

    with XYZTrajectoryFile(get_fn('frame0.xyz')) as f:
        f.seek(1)
        eq(1, f.tell())
        xyz1 = f.read(n_frames=1)
        eq(reference.xyz[1], xyz1[0] / 10)

        f.seek(10)
        eq(10, f.tell())
        xyz10 = f.read(n_frames=1)
        eq(reference.xyz[10], xyz10[0] / 10)
        eq(11, f.tell())

        f.seek(-8, 1)
        xyz3 = f.read(n_frames=1)
        eq(reference.xyz[3], xyz3[0] / 10)

        f.seek(4, 1)
        xyz8 = f.read(n_frames=1)
        eq(reference.xyz[8], xyz8[0] / 10)


def test_len(get_fn):
    with md.open(get_fn('frame0.xyz')) as fh:
        assert len(fh) == 501
        assert fh._frame_index == 0
        assert len(fh.read()) == 501



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

from mdtraj.testing import eq
from mdtraj import load, load_topology


def test_xml(get_fn):
    top = load_topology(get_fn('native2.pdb'), no_boxchk=True)
    t1 = load(get_fn('native2.xml'), top=top)
    t2 = load(get_fn('native2.pdb'), no_boxchk=True)

    t1.center_coordinates()
    t2.center_coordinates()

    assert eq(t1.xyz, t2.xyz)
    assert eq(t1.unitcell_vectors, t2.unitcell_vectors)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

from mdtraj.formats import TRRTrajectoryFile
import os, tempfile
import numpy as np
from mdtraj.testing import eq
from mdtraj import io
import pytest

fd, temp = tempfile.mkstemp(suffix='.trr')
os.close(fd)


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.unlink(temp)


def test_1():
    # Write data and read it back
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    time = np.random.randn(500)
    step = np.arange(500)
    lambd = np.random.randn(500)

    with TRRTrajectoryFile(temp, 'w') as f:
        f.write(xyz=xyz, time=time, step=step, lambd=lambd)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2 = f.read()

    assert eq(xyz, xyz2)
    assert eq(time, time2)
    assert eq(step, step2)
    assert eq(lambd, lambd2)


def test_read_stride(get_fn):
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz, time, step, box, lambd = f.read()
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz3, time3, step3, box3, lambd3 = f.read(stride=3)
    assert eq(xyz[::3], xyz3)
    assert eq(step[::3], step3)
    assert eq(box[::3], box3)
    assert eq(time[::3], time3)


def test_read_stride_n_frames(get_fn):
    # trr read stride when n_frames is supplied (different path)
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz, time, step, box, lambd = f.read()
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz3, time3, step3, box3, lambd3 = f.read(n_frames=1000, stride=3)
    assert eq(xyz[::3], xyz3)
    assert eq(step[::3], step3)
    assert eq(box[::3], box3)
    assert eq(time[::3], time3)


def test_read_stride_offsets(get_fn):
    # read xtc with stride and offsets
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz, time, step, box, lambd = f.read()
    for s in (1, 2, 3, 4, 5):
        with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
            f.offsets # pre-compute byte offsets between frames
            xyz_s, time_s, step_s, box_s, lamb_s = f.read(stride=s)
        assert eq(xyz_s, xyz[::s])
        assert eq(step_s, step[::s])
        assert eq(box_s, box[::s])
        assert eq(time_s, time[::s])


def test_read_stride_n_frames_offsets(get_fn):
    # read trr with stride with n_frames and offsets
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz, time, step, box, lambd = f.read()
    for s in (1, 2, 3, 4, 5):
        with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
            f.offsets # pre-compute byte offsets between frames
            xyz_s, time_s, step_s, box_s, lamb_s = f.read(n_frames=1000, stride=s)
        assert eq(xyz_s, xyz[::s], err_msg='stride=%s' % s)
        assert eq(step_s, step[::s])
        assert eq(box_s, box[::s])
        assert eq(time_s, time[::s])


def test_read_stride_switching(get_fn):
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        xyz, time, step, box, lambd = f.read()
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        f.offsets  # pre-compute byte offsets between frames
        # read the first 10 frames with stride of 2
        s = 2
        n_frames = 10
        xyz_s, time_s, step_s, box_s, lamb_s = f.read(n_frames=n_frames, stride=s)
        assert eq(xyz_s, xyz[:n_frames*s:s])
        assert eq(step_s, step[:n_frames*s:s])
        assert eq(box_s, box[:n_frames*s:s])
        assert eq(time_s, time[:n_frames*s:s])
        # now read the rest with stride 3, should start from frame index 8.
        # eg. np.arange(0, n_frames*s + 1, 2)[-1] == 18
        offset = f.tell()
        assert offset == 20
        s = 3
        xyz_s, time_s, step_s, box_s, lamb_s = f.read(n_frames=None, stride=s)
        assert eq(xyz_s, xyz[offset::s])
        assert eq(step_s, step[offset::s])
        assert eq(box_s, box[offset::s])
        assert eq(time_s, time[offset::s])


def test_write_read():
    # Write data and read it back
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    time = np.random.randn(500)
    step = np.arange(500)
    lambd = np.random.randn(500)

    with TRRTrajectoryFile(temp, 'w') as f:
        f.write(xyz=xyz, time=time, step=step, lambd=lambd)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2 = f.read(n_frames=500)

    assert eq(xyz, xyz2)
    assert eq(time, time2)
    assert eq(step, step2)
    assert eq(lambd, lambd2)


def test_read_atomindices_1():
    with TRRTrajectoryFile(temp) as f:
        xyz, time, step, box, lambd = f.read()

    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2 = f.read(atom_indices=[0, 1, 2])
    assert eq(xyz[:, [0, 1, 2]], xyz2)
    assert eq(step, step2)
    assert eq(box, box2)
    assert eq(lambd, lambd2)
    assert eq(time, time2)


def test_read_atomindices_2():
    with TRRTrajectoryFile(temp) as f:
        xyz, time, step, box, lambd = f.read()

    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2 = f.read(atom_indices=slice(None, None, 2))
    assert eq(xyz[:, ::2], xyz2)
    assert eq(step, step2)
    assert eq(box, box2)
    assert eq(lambd, lambd2)
    assert eq(time, time2)


def test_deficient_shape():
    # Write data one frame at a time. This checks how the shape is dealt with,
    # because each frame is deficient in shape.
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    time = np.random.randn(500)
    step = np.arange(500)
    lambd = np.random.randn(500)

    with TRRTrajectoryFile(temp, 'w') as f:
        for i in range(len(xyz)):
            f.write(xyz=xyz[i], time=time[i], step=step[i], lambd=lambd[i])
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2 = f.read()

    assert eq(xyz, xyz2)
    assert eq(time, time2)
    assert eq(step, step2)
    assert eq(lambd, lambd2)


def test_ragged_1():
    # try first writing no box vectors,, and then adding some
    xyz = np.random.randn(100, 5, 3)
    time = np.random.randn(100)
    box = np.random.randn(100, 3, 3)

    with TRRTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz)
        with pytest.raises(ValueError):
            f.write(xyz, time, box)


def test_ragged_2():
    # try first writing no box vectors, and then adding some
    xyz = np.random.randn(100, 5, 3)
    time = np.random.randn(100)
    box = np.random.randn(100, 3, 3)

    with TRRTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time=time, box=box)
        with pytest.raises(ValueError):
            f.write(xyz)


def test_malformed():
    with open(temp, 'w') as tmpf:
        tmpf.write("foo")  # very badly malformed TRR

    with pytest.raises(IOError):
        TRRTrajectoryFile(temp)

    psutil = pytest.importorskip("psutil")
    open_files = psutil.Process().open_files()
    paths = [os.path.realpath(f.path) for f in open_files]
    assert os.path.realpath(temp) not in paths


def test_tell(get_fn):
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        eq(f.tell(), 0)

        f.read(101)
        eq(f.tell(), 101)

        f.read(3)
        eq(f.tell(), 104)


def test_seek(get_fn):
    reference = TRRTrajectoryFile(get_fn('frame0.trr')).read()[0]
    with TRRTrajectoryFile(get_fn('frame0.trr')) as f:
        eq(len(f), len(reference))
        eq(len(f.offsets), len(reference))

        eq(f.tell(), 0)
        eq(f.read(1)[0][0], reference[0])
        eq(f.tell(), 1)

        xyz = f.read(1)[0][0]
        eq(xyz, reference[1])
        eq(f.tell(), 2)

        f.seek(0)
        eq(f.tell(), 0)
        xyz = f.read(1)[0][0]
        eq(f.tell(), 1)
        eq(xyz, reference[0])

        f.seek(5)
        eq(f.read(1)[0][0], reference[5])
        eq(f.tell(), 6)

        f.seek(-5, 1)
        eq(f.tell(), 1)
        eq(f.read(1)[0][0], reference[1])


def test_get_velocities():
    """Write data with velocities and read it back"""
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    vel = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f._write(xyz=xyz, time=time, step=step, box=box, lambd=lambd, vel=vel)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2, vel2, forces2 = f._read(
            n_frames=500, atom_indices=None, get_velocities=True,
            get_forces=False
        )

    eq(xyz, xyz2)
    eq(time, time2)
    eq(step, step2)
    eq(lambd, lambd2)
    eq(vel, vel2)
    eq(None, forces2)


def test_get_forces():
    """Write data with forces and read it back"""
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    forces = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f._write(xyz=xyz, time=time, step=step, box=box, lambd=lambd,
                 forces=forces)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2, vel2, forces2 = f._read(
            n_frames=500, atom_indices=None, get_velocities=False,
            get_forces=True
        )

    eq(xyz, xyz2)
    eq(time, time2)
    eq(step, step2)
    eq(lambd, lambd2)
    eq(None, vel2)
    eq(forces, forces2)


def test_get_velocities_and_forces():
    """Write data with velocities and forces, and read it back"""
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    vel = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    forces = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f._write(xyz=xyz, time=time, step=step, box=box, lambd=lambd,
                 vel=vel, forces=forces)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2, vel2, forces2 = f._read(
            n_frames=500, atom_indices=None, get_velocities=True,
            get_forces=True
        )

    eq(xyz, xyz2)
    eq(vel, vel2)
    eq(forces, forces2)
    eq(time, time2)
    eq(step, step2)
    eq(lambd, lambd2)


def test_get_velocities_forces_atom_indices_1():
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    vel = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    forces = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f._write(xyz=xyz, time=time, step=step, box=box, lambd=lambd,
                 vel=vel, forces=forces)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2, vel2, forces2 = f._read(
            n_frames=500, atom_indices=[0, 1, 2],
            get_velocities=True, get_forces=True
        )

    eq(xyz[:, [0, 1, 2]], xyz2)
    eq(vel[:, [0, 1, 2]], vel2)
    eq(forces[:, [0, 1, 2]], forces2)
    eq(time, time2)
    eq(step, step2)
    eq(lambd, lambd2)


def test_get_velocities_forces_atom_indices_2():
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    vel = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    forces = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f._write(xyz=xyz, time=time, step=step, box=box, lambd=lambd,
                 vel=vel, forces=forces)
    with TRRTrajectoryFile(temp) as f:
        xyz2, time2, step2, box2, lambd2, vel2, forces2 = f._read(
            n_frames=500, atom_indices=slice(None, None, 2),
            get_velocities=True, get_forces=True
        )

    eq(xyz[:, ::2], xyz2)
    eq(vel[:, ::2], vel2)
    eq(forces[:, ::2], forces2)
    eq(time, time2)
    eq(step, step2)
    eq(lambd, lambd2)
    pass


def test_read_velocities_do_not_exist():
    """Requesting velocities from a file that does not have them"""
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f.write(xyz=xyz, time=time, step=step, box=box, lambd=lambd)
    with TRRTrajectoryFile(temp) as f:
        with pytest.raises(RuntimeError):
            f._read(n_frames=500, atom_indices=None, get_velocities=True, get_forces=False)


def test_read_forces_do_not_exist():
    """Requesting forces from a file that does not have them"""
    # NOTE: this is a test of a hidden API
    xyz = np.array(np.random.randn(500, 50, 3), dtype=np.float32)
    box = np.array(np.random.randn(500, 3, 3), dtype=np.float32)
    time = np.array(np.random.randn(500), dtype=np.float32)
    step = np.array(np.arange(500), dtype=np.int32)
    lambd = np.array(np.random.randn(500), dtype=np.float32)

    with TRRTrajectoryFile(temp, 'w') as f:
        f.write(xyz=xyz, time=time, step=step, box=box, lambd=lambd)
    with TRRTrajectoryFile(temp) as f:
        with pytest.raises(RuntimeError):
            f._read(n_frames=500, atom_indices=None, get_velocities=False, get_forces=True)



from __future__ import print_function, absolute_import, division

import gzip
import bz2
from mdtraj.testing import eq
from mdtraj.utils import open_maybe_zipped


def test_read_gz(tmpdir):
    fn = '{}/read.gz'.format(tmpdir)
    with gzip.GzipFile(fn, 'w') as f:
        f.write('COOKIE'.encode('utf-8'))
    eq(open_maybe_zipped(fn, 'r').read(), u'COOKIE')


def test_write_gz(tmpdir):
    fn = '{}/write.gz'.format(tmpdir)
    with open_maybe_zipped(fn, 'w') as f:
        f.write(u'COOKIE')
    with gzip.GzipFile(fn, 'r') as f:
        eq(f.read().decode('utf-8'), u'COOKIE')


def test_read_bz2(tmpdir):
    fn = '{}/read.bz2'.format(tmpdir)
    with bz2.BZ2File(fn, 'w') as f:
        f.write('COOKIE'.encode('utf-8'))
    eq(open_maybe_zipped(fn, 'r').read(), u'COOKIE')


def test_write_bz2(tmpdir):
    fn = '{}/write.bz2'.format(tmpdir)
    with open_maybe_zipped(fn, 'w') as f:
        f.write(u'COOKIE')
    with bz2.BZ2File(fn, 'r') as f:
        eq(f.read().decode('utf-8'), u'COOKIE')



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import time
import itertools
import pytest
import numpy as np
import mdtraj as md

from mdtraj.testing import eq, assert_allclose
from mdtraj.geometry.distance import compute_distances, compute_distances_t, compute_displacements, find_closest_contact
from mdtraj.geometry.distance import _displacement_mic, _displacement

N_FRAMES = 20
N_ATOMS = 20

xyz = np.asarray(np.random.randn(N_FRAMES, N_ATOMS, 3), dtype=np.float32)
pairs = np.array(list(itertools.combinations(range(N_ATOMS), 2)), dtype=np.int32)
times = np.array([[i, 0] for i in range(N_FRAMES)[::2]], dtype=np.int32)

ptraj = md.Trajectory(xyz=xyz, topology=None)
ptraj.unitcell_vectors = np.ascontiguousarray(np.random.randn(N_FRAMES, 3, 3) + 2 * np.eye(3, 3), dtype=np.float32)


def test_generator():
    pairs2 = itertools.combinations(range(N_ATOMS), 2)
    a = compute_distances(ptraj, pairs)
    b = compute_distances(ptraj, pairs2)
    eq(a, b)


def test_0():
    a = compute_distances(ptraj, pairs, periodic=False, opt=True)
    b = compute_distances(ptraj, pairs, periodic=False, opt=False)
    eq(a, b)


def test_1():
    a = compute_displacements(ptraj, pairs, periodic=False, opt=True)
    b = compute_displacements(ptraj, pairs, periodic=False, opt=False)
    eq(a, b)


def test_2():
    a = compute_distances(ptraj, pairs, periodic=False, opt=False)
    b = compute_displacements(ptraj, pairs, periodic=False, opt=False)
    eq(a, np.sqrt(np.sum(np.square(b), axis=2)))


def test_3():
    a = compute_distances(ptraj, pairs, periodic=False, opt=True)
    b = compute_displacements(ptraj, pairs, periodic=False, opt=True)
    eq(a, np.sqrt(np.sum(np.square(b), axis=2)))


def test_0p():
    a = compute_distances(ptraj, pairs, periodic=True, opt=True)
    b = compute_distances(ptraj, pairs, periodic=True, opt=False)
    print(a, b)
    eq(a, b, decimal=3)


def test_1p():
    a = compute_displacements(ptraj, pairs, periodic=True, opt=True)
    b = compute_displacements(ptraj, pairs, periodic=True, opt=False)
    eq(a, b, decimal=3)


def test_2p():
    a = compute_distances(ptraj, pairs, periodic=True, opt=False)
    b = compute_displacements(ptraj, pairs, periodic=True, opt=False)
    assert a.shape == (len(ptraj), len(pairs))
    assert b.shape == (len(ptraj), len(pairs), 3), str(b.shape)
    b = np.sqrt(np.sum(np.square(b), axis=2))
    eq(a, b, decimal=5)


def test_3p():
    a = compute_distances(ptraj, pairs, periodic=True, opt=True)
    b = compute_displacements(ptraj, pairs, periodic=True, opt=True)
    print(a,  b)
    eq(a, np.sqrt(np.sum(np.square(b), axis=2)))


def test_4():
    # using a really big box, we should get the same results with and without
    # pbcs
    box = np.array([[100, 0, 0], [0, 200, 0], [0, 0, 300]])
    box = np.zeros((N_FRAMES, 3, 3)) + box  # broadcast it out
    a = _displacement_mic(xyz, pairs, box, False)
    b = _displacement(xyz, pairs)
    eq(a, b, decimal=3)


def test_5():
    # simple wrap around along the z axis.
    xyz = np.array([[[0.0, 0.0, 0.0], [0.0, 0.0, 2.2]]])
    box = np.eye(3, 3).reshape(1, 3, 3)
    result = _displacement_mic(xyz, np.array([[0, 1]]), box, True)
    eq(result, np.array([[[0, 0, 0.2]]]))


def test_6(get_fn):
    ext_ref = np.array([17.4835, 22.2418, 24.2910, 22.5505, 12.8686, 22.1090,
                        7.4472, 22.4253, 19.8283, 20.6935]) / 10
    traj = md.load(get_fn('test_good.nc'), top=get_fn('test.parm7'))
    _run_amber_traj(traj, ext_ref)


def test_7(get_fn):
    ext_ref = np.array([30.9184, 23.9040, 25.3869, 28.0060, 25.9704, 24.6836,
                        23.0508, 27.1983, 24.4954, 26.7448]) / 10
    traj = md.load(get_fn('test_bad.nc'), top=get_fn('test.parm7'))
    _run_amber_traj(traj, ext_ref)


def _run_amber_traj(traj, ext_ref):
    # Test triclinic case where simple approach in Tuckerman text does not
    # always work
    distopt = md.compute_distances(traj, [[0, 9999]], opt=True)
    distslw = md.compute_distances(traj, [[0, 9999]], opt=False)
    dispopt = md.compute_displacements(traj, [[0, 9999]], opt=True)
    dispslw = md.compute_displacements(traj, [[0, 9999]], opt=False)

    eq(distopt, distslw, decimal=5)
    eq(dispopt, dispslw, decimal=5)

    assert_allclose(distopt.flatten(), ext_ref, atol=2e-5)

    # Make sure distances from displacements are the same
    eq(np.sqrt((dispopt.squeeze() ** 2).sum(axis=1)), distopt.squeeze())
    eq(np.sqrt((dispslw.squeeze() ** 2).sum(axis=1)), distslw.squeeze())
    eq(dispopt, dispslw, decimal=5)


def test_closest_contact():
    box_size = np.array([3.0, 4.0, 5.0])
    traj = md.Trajectory(xyz=xyz * box_size, topology=None)
    _verify_closest_contact(traj)
    traj.unitcell_lengths = np.array([box_size for i in range(N_FRAMES)])
    traj.unitcell_angles = np.array([[90.0, 90.0, 90.0] for i in range(N_FRAMES)])
    _verify_closest_contact(traj)
    traj.unitcell_angles = np.array([[80.0, 90.0, 100.0] for i in range(N_FRAMES)])
    _verify_closest_contact(traj)


def _verify_closest_contact(traj):
    group1 = np.array([i for i in range(N_ATOMS // 2)], dtype=int)
    group2 = np.array([i for i in range(N_ATOMS // 2, N_ATOMS)], dtype=int)
    contact = find_closest_contact(traj, group1, group2)
    pairs = np.array([(i, j) for i in group1 for j in group2], dtype=int)
    dists = md.compute_distances(traj, pairs, True)[0]
    dists2 = md.compute_distances(traj, pairs, False)[0]
    nearest = np.argmin(dists)
    eq(float(dists[nearest]), contact[2], decimal=5)
    assert ((pairs[nearest, 0] == contact[0] and pairs[nearest, 1] == contact[1]) or (
    pairs[nearest, 0] == contact[1] and pairs[nearest, 1] == contact[0]))


def test_distance_nan():
    xyz = np.array([[1, 1, 1], [2, 1, 1], [np.nan, np.nan, np.nan]]).reshape(1, 3, 3)
    dists = md.compute_distances(md.Trajectory(xyz=xyz, topology=None), [[0, 1]])
    assert np.isfinite(dists).all()


def test_closest_contact_nan_pos():
    box_size = np.array([3.0, 4.0, 5.0])
    xyz = np.asarray(np.random.randn(2, 20, 3), dtype=np.float32)
    xyz *= box_size
    # Set the last frame to nan
    xyz[-1] = np.nan
    # Slice of the last frame, so nans should not cause troubles.
    xyz = xyz[:-1]
    traj = md.Trajectory(xyz=xyz, topology=None)
    _verify_closest_contact(traj)


def test_distance_t_inputs():
    incorrect_pairs = np.array((0, ptraj.n_atoms+1))
    with pytest.raises(ValueError, match='atom_pairs'):
        compute_distances_t(ptraj, incorrect_pairs, times)

    incorrect_times = np.array((0, ptraj.n_frames+1))
    with pytest.raises(ValueError, match='time_pairs'):
        compute_distances_t(ptraj, pairs, incorrect_times)


def test_distances_t(get_fn):
    a = compute_distances_t(ptraj, pairs, times, periodic=True, opt=True)
    b = compute_distances_t(ptraj, pairs, times, periodic=True, opt=False)
    eq(a, b)
    c = compute_distances_t(ptraj, pairs, times, periodic=False, opt=True)
    d = compute_distances_t(ptraj, pairs, times, periodic=False, opt=False)
    eq(c, d)

def test_distances_t_at_0(get_fn):
    times = np.array([[0, 0]], dtype=np.int32)
    a = compute_distances_t(ptraj, pairs, times, periodic=True, opt=True)
    b = compute_distances_t(ptraj, pairs, times, periodic=True, opt=False)
    c = compute_distances(ptraj[:1], pairs, periodic=True, opt=True)
    d = compute_distances(ptraj[:1], pairs, periodic=True, opt=False)
    eq(a, c)
    eq(b, d)

def _run_amber_traj_t(traj, ext_ref):
    # Test triclinic case where simple approach in Tuckerman text does not
    # always work
    distopt = compute_distances_t(traj, atom_pairs=[[0, 9999]], time_pairs=[[0, 2]], opt=True)
    distslw = compute_distances_t(traj, atom_pairs=[[0, 9999]], time_pairs=[[0, 2]], opt=False)

def test_amber_t(get_fn):
    ext_ref = np.array([17.4835, 22.2418, 24.2910, 22.5505, 12.8686, 22.1090,
                        7.4472, 22.4253, 19.8283, 20.6935]) / 10
    traj = md.load(get_fn('test_good.nc'), top=get_fn('test.parm7'))
    _run_amber_traj_t(traj, ext_ref)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################
from __future__ import print_function, division
import sys
import types
import warnings
import pkgutil
import importlib
from inspect import (isclass, ismodule, isfunction, ismethod,
                     getmembers, getdoc, getmodule, getargs, isbuiltin)
from mdtraj.testing.docscrape import NumpyDocString
from mdtraj.utils.six import get_function_code, get_function_closure, PY2

__all__ = ['docstring_verifiers', 'import_all_modules']

##############################################################################
# functions
##############################################################################


def docstring_verifiers(module, error_on_none=False):
    """Yield an iterable of tests that verify the docstrings (for
    adherance to the NumpyDoc format) of all functions/classes defined
    in a module.

    Parameters
    ----------
    module : module
        The module to test
    error_on_none : bool, default=False
        Throw an error if no docstring is defined
    """

    # These are the types that we want to check
    # currently, the docstring on classes are not being checked, since
    # isclass() is not in the list
    acceptors = [isfunction, ismethod, isbuiltin]
    accept = lambda f: any([acc(f) for acc in acceptors])
    functions = [f for f in walk(module) if accept(f)]

    def format(f):
        """
        Format a method/function/class as a string

        Parameters
        ----------
        f : function, method, class

        Returns
        -------
        repr : string
            A string represntation
        """
        if ismethod(f):
            if PY2:
                return '.'.join([getmodule(f).__name__, f.im_class.__name__, f.__name__])
            else:
                return '.'.join([getmodule(f).__name__, f.__self__.__class__.__name__, f.__name__])
        if isfunction(f) or isbuiltin(f):
            return '.'.join([getmodule(f).__name__, f.__name__])
        if isclass(f):
            return f.__name__
        return 'Error'

    def check_docstring(f):
        """
        Ensure the docstring of `f` is in accordance with the numpy standard

        Currently, only the Parameters section of the docstring is checked.

        Parameters
        ----------
        f : function, method, class

        Returns
        -------
        repr : string
            A string represntation
        """

        doc = getdoc(f)
        if doc is None:
            if error_on_none:
                raise ValueError('no docstring for %s' % format(f))
        else:
            with warnings.catch_warnings():
                warnings.simplefilter('error')
                parsed = NumpyDocString(doc)

            param_names = {e[0] for e in parsed['Parameters']}

            if isbuiltin(f):
                # You can't get the arglist from a builtin, which
                # is how cython functions turn up

                # but you can, hackily, get the number of arguments it wants
                # by parseing the error hen you supply too many
                import re
                try:
                    f(*list(range(100)))
                except TypeError as e:
                    m = re.search('takes at most (\d+) positional arguments', str(e))
                    if not m:
                        return
                    n_args = int(m.group(1))

                if len(param_names) != n_args:
                    raise ValueError("In %s, number of arguments, %d, doesn't "
                        " match the length of the Parameters in the "
                        "docstring, %d" % (format(f), n_args, len(param_names)))
                return

            args = set(getargs(get_function_code(f)).args)
            if 'self' in args:
                args.remove('self')
            if 'cls' in args:
                args.remove('cls')

            if args != param_names:
                raise ValueError("In %s, arguments %s don't "
                    "match Parameters list %s" % (format(f),
                        list(args), list(param_names)))

    for f in functions:
        qq = lambda: check_docstring(f)
        qq.description = 'NumpyDoc: %s.%s' % (module.__name__, f.__name__)
        qq.fname = f.__name__
        yield qq


def ispackage(obj):
    """
    Check if obj is a package. Simply look for whether its a module whose
    filename is __init__.py(c)

    Parameters
    ----------
    obj : module
    """
    if ismodule(obj):
        return obj.__file__.endswith("__init__.pyc") or \
            obj.__file__.endswith("__init__.py")
    return False


def walk(module):
    """
    Get all of the functions, classes and their methods defined within
    a python module

    Parameters
    ----------
    module : module

    Returns
    -------
    funcs : list
        List of functions, classes and methods
    """
    assert ismodule(module)
    if ispackage(module):
        raise ValueError('Sorry, you need to supply me a module, not a package')

    def is_valid(obj):
        if getmodule(obj) == module:
            # cython specific stuff
            if module.__file__.endswith('.so'):
                if isbuiltin(obj):
                    return not obj.__name__.startswith('_')

            if ismethod(obj) or isfunction(obj):
                return not obj.__name__.startswith('_')
            if isclass(obj):
                return True
        return False

    instack = [v for k, v in getmembers(module) if is_valid(v)]
    outstack = []

    while True:
        try:
            item = instack.pop()
        except IndexError:
            break
        outstack.append(item)

        if isclass(item):
            instack.extend([v for k, v in getmembers(item) if is_valid(v)])

    return outstack


def import_all_modules(pkg):
    result = []
    for _, modname, ispkg in pkgutil.iter_modules(pkg.__path__):
        c = '%s.%s' % (pkg.__name__, modname)
        if modname.startswith('test_'):
            continue
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                mod = importlib.import_module(c)
            if ispkg:
                result.extend(import_all_modules(mod))
            else:
                result.append(mod)
        except ImportError as e:
            print('e', e)
            continue

    return result



import mdtraj as md
from mdtraj.formats import TNGTrajectoryFile
from mdtraj.testing import eq
import os
import tempfile
import numpy as np


def test_load_trajectory(get_fn):
    # Compare a TNG file to the PDB file it was created from.

    pdbtraj = md.load_pdb(get_fn('frame0.pdb'))
    tngtraj = md.load_tng(get_fn('frame0.tng'), top=pdbtraj.topology)
    eq(pdbtraj.n_frames, tngtraj.n_frames)
    eq(pdbtraj.unitcell_vectors, tngtraj.unitcell_vectors)
    eq(pdbtraj.xyz, tngtraj.xyz)


def test_load_topology(get_fn):
    # Test loading a Topology from a TNG file.

    traj = md.load_tng(get_fn('tng_example.tng'))
    top = traj.topology
    eq(5, top.n_residues)
    eq(10, top.n_bonds)
    bonds = [(a, b) for a, b in top.bonds]
    for res in top.residues:
        eq('HOH', res.name)
        eq(3, res.n_atoms)
        eq(md.element.oxygen, res.atom(0).element)
        eq(md.element.hydrogen, res.atom(1).element)
        eq(md.element.hydrogen, res.atom(2).element)
        assert (res.atom(0), res.atom(1)) in bonds
        assert (res.atom(0), res.atom(2)) in bonds


def test_write(tmpdir):
    # Write a TNG file, then read it back.
    xyz = np.asarray(np.around(np.random.randn(100, 10, 3), 3), dtype=np.float32)
    time = 0.1 * np.arange(0, 100, dtype=np.float32)
    box = np.asarray(np.random.randn(100, 3, 3), dtype=np.float32)

    outfn = "{}/tmp.tng".format(tmpdir)
    with TNGTrajectoryFile(outfn, 'w') as f:
        f.write(xyz, time=time, box=box)
    with TNGTrajectoryFile(outfn) as f:
        xyz2, time2, box2 = f.read()
    eq(xyz, xyz2)
    eq(time, time2)
    eq(box, box2)



import itertools
import mdtraj as md
from mdtraj.testing import eq
import numpy as np


def test_angle_pbc_0():
    # Check that angles are calculated correctly across periodic boxes
    random = np.random.RandomState(0)
    boxLengths = [1, 2, 3]
    N = 1000

    X = np.cumsum(0.1 * random.randn(N * 3, 3), axis=0).reshape(N, 3, 3)

    center = X.mean(axis=1).reshape(N, 1, 3)
    cell = np.floor(center / boxLengths)
    X_cell = X - (cell * boxLengths)

    assert not np.all(X == X_cell)

    t0 = md.Trajectory(xyz=X, topology=None)
    t1 = md.Trajectory(xyz=X_cell, topology=None)
    t1.unitcell_vectors = np.tile(np.diag(boxLengths), [len(X), 1, 1])

    r0 = md.compute_angles(t0, [[0, 1, 2]], opt=False).reshape(-1)
    r1 = md.compute_angles(t1, [[0, 1, 2]], opt=False).reshape(-1)
    r2 = md.compute_angles(t0, [[0, 1, 2]], opt=True).reshape(-1)
    r3 = md.compute_angles(t1, [[0, 1, 2]], opt=True).reshape(-1)

    np.testing.assert_array_almost_equal(r0, r1, decimal=4)
    np.testing.assert_array_almost_equal(r2, r3, decimal=4)
    np.testing.assert_array_almost_equal(r0, r3, decimal=4)


def test_generator():
    N_FRAMES = 2
    N_ATOMS = 5
    xyz = np.asarray(np.random.randn(N_FRAMES, N_ATOMS, 3), dtype=np.float32)
    ptraj = md.Trajectory(xyz=xyz, topology=None)

    triplets = np.array(list(itertools.combinations(range(N_ATOMS), 3)), dtype=np.int32)
    triplets2 = itertools.combinations(range(N_ATOMS), 3)
    a = md.compute_angles(ptraj, triplets)
    b = md.compute_angles(ptraj, triplets2)
    eq(a, b)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import mdtraj as md
from mdtraj.testing import eq
from mdtraj.formats import mol2
from distutils.spawn import find_executable
import tarfile
import pickle
import os
import numpy as np
import scipy.sparse
import pytest


def test_load_mol2(get_fn):
    trj = md.load(get_fn('imatinib.mol2'))
    ref_trj = md.load(get_fn('imatinib.pdb'))
    eq(trj.xyz, ref_trj.xyz)

    ref_top, ref_bonds = ref_trj.top.to_dataframe()
    top, bonds = trj.top.to_dataframe()
    # PDB Does not have bond order, ensure that the equality fails
    try:
        eq(bonds, ref_bonds)
    except AssertionError:
        # This is what we wanted to happen, its fine
        pass
    else:
        raise AssertionError("Reference bonds with no bond order should not equal Mol2 bonds with bond order")
    # Strip bond order info since PDB does not have it
    bonds[:, -2:] = np.zeros([bonds.shape[0], 2])
    eq(bonds, ref_bonds)


@pytest.mark.skipif(find_executable('obabel') is None, reason='Requires obabel')
@pytest.mark.skipif(os.environ.get("TRAVIS", None) == 'true', reason="Skip on Travis.")
def test_load_freesolv_gaffmol2_vs_sybylmol2_vs_obabelpdb(get_fn, tmpdir):
    tar_filename = "freesolve_v0.3.tar.bz2"
    tar = tarfile.open(get_fn(tar_filename), mode="r:bz2")
    os.chdir(str(tmpdir))
    tar.extractall()
    tar.close()

    with open("./v0.3/database.pickle", 'rb') as f:
        database = pickle.load(f, encoding='latin-1')

    for key in database:
        gaff_filename = "./v0.3/mol2files_gaff/%s.mol2" % key
        pdb_filename = "./v0.3/mol2files_gaff/%s.pdb" % key
        sybyl_filename = "./v0.3/mol2files_sybyl/%s.mol2" % key

        cmd = "obabel -imol2 %s -opdb > %s 2>/dev/null" % (sybyl_filename, pdb_filename)
        assert os.system(cmd) == 0

        t_pdb = md.load(pdb_filename)
        t_gaff = md.load(gaff_filename)
        t_sybyl = md.load(sybyl_filename)

        eq(t_pdb.n_atoms, t_gaff.n_atoms)
        eq(t_pdb.n_atoms, t_sybyl.n_atoms)

        eq(t_pdb.n_frames, t_gaff.n_frames)
        eq(t_pdb.n_frames, t_gaff.n_frames)

        eq(t_pdb.xyz, t_gaff.xyz, decimal=4)
        eq(t_pdb.xyz, t_sybyl.xyz, decimal=4)

        top_pdb, bonds_pdb = t_pdb.top.to_dataframe()
        top_gaff, bonds_gaff = t_gaff.top.to_dataframe()
        top_sybyl, bonds_sybyl = t_sybyl.top.to_dataframe()

        eq(top_sybyl.name.values, top_pdb.name.values)

        # eq(top_gaff.name.values, top_sybyl.name.values)  # THEY CAN HAVE DIFFERENT NAMES, so this isn't TRUE!

        def make_bonds_comparable(bond_array):
            """Create a bond connectivity matrix from a numpy array of atom pairs.  Avoids having to compare the order in which bonds are listed."""
            n_bonds = len(bond_array)
            data = np.ones(n_bonds)
            i = bond_array[:, 0]
            j = bond_array[:, 1]
            matrix = scipy.sparse.coo_matrix((data, (i, j)), shape=(t_pdb.n_atoms, t_pdb.n_atoms)).toarray()
            return matrix + matrix.T  # Symmetrize to account for (a ~ b) versus (b ~ a)

        bond_matrix_pdb = make_bonds_comparable(bonds_pdb)
        bond_matrix_gaff = make_bonds_comparable(bonds_gaff)
        bond_matrix_sybyl = make_bonds_comparable(bonds_sybyl)

        eq(bond_matrix_pdb, bond_matrix_gaff)
        eq(bond_matrix_pdb, bond_matrix_sybyl)

        # Third row from mol2 file copied below, used in testing.
        #       3 N1          8.5150   -0.1620    1.3310 n3        1 LIG     -0.732600


def test_mol2_dataframe(get_fn):
    top, bonds = mol2.mol2_to_dataframes(get_fn("imatinib.mol2"))
    eq(top.name[2], "N1")
    eq(top.atype[2], "n3")
    eq(top.resName[2], "LIG")
    eq(float(top.charge[2]), -0.732600)


def test_mol2_dataframe_status(get_fn):
    atoms, bonds = mol2.mol2_to_dataframes(get_fn('adp.mol2'))
    assert atoms['charge'][1] == 1.3672
    assert atoms['status'][1] == '****'


def test_mol2_warnings(get_fn):
    trj = md.load_mol2(get_fn('lysozyme-ligand-tripos.mol2'))


def test_mol2_status_bits(get_fn):
    trj = md.load_mol2(get_fn('status-bits.mol2'))
    eq(trj.topology.n_atoms, 18)
    eq(trj.topology.n_bonds, 18)


def test_mol2_without_bonds(get_fn):
    trj = md.load_mol2(get_fn('li.mol2'))
    assert trj.topology.n_bonds == 0



def test_mol2_element_name(get_fn):
    trj = md.load_mol2(get_fn('cl.mol2'))
    top, bonds = trj.top.to_dataframe()
    assert top.iloc[0]['element'] == 'Cl'

    
@pytest.mark.parametrize('mol2_file', [('li.mol2'),
('lysozyme-ligand-tripos.mol2'), ('imatinib.mol2'),
('status-bits.mol2'), ('adp.mol2'), ('water_acn.mol2')])
def test_load_all_mol2(mol2_file, get_fn):
    trj = md.load_mol2(get_fn(mol2_file))

def test_mol2_n_residues(get_fn):
    trj = md.load_mol2(get_fn('water_acn.mol2'))
    assert trj.n_residues == 10




cdef extern from "include/dcdplugin.h":
    ctypedef struct dcdhandle:
      long fd
      int natoms
      int nsets
      int setsread
      int istart
      int nsavc
      double delta
      int nfixed
      float *x, *y, *z
      int *freeind
      float *fixedcoords
      int reverse
      int charmm
      int first
      int with_unitcell

    dcdhandle* open_dcd_read(char *path, char *filetype, int *natoms, int* nsets)
    int read_next_timestep(dcdhandle *v, int natoms, molfile_timestep_t *ts)
    void close_file_read(dcdhandle *v)

    dcdhandle* open_dcd_write(const char *path, const char *filetype, const int natoms, const int with_unitcell)
    int write_timestep(dcdhandle *v, molfile_timestep_t *ts)
    void close_file_write(dcdhandle *v)
    int dcd_nsets(dcdhandle* v)
    int dcd_rewind(dcdhandle* dcd)


cdef extern from "include/molfile_plugin.h":
    ctypedef struct molfile_timestep_t:
      float *coords  # coordinates of all atoms, arranged xyzxyzxyz
      float *velocities  # space for velocities of all atoms; same layout
      float A, B, C, alpha, beta, gamma




# Copied from the yt_project, commit e8fb57e
# yt/doc/extensions/notebook_sphinxext.py
#  https://bitbucket.org/yt_analysis/yt/src/e8fb57e66ca42e26052dadf054a5c782740abec9/doc/extensions/notebook_sphinxext.py?at=yt

# Almost completely re-written by Matthew Harrigan to use nbconvert v4

from __future__ import print_function

import os
import shutil

from docutils import nodes
from docutils.parsers.rst import directives, Directive
import nbformat
from nbconvert import HTMLExporter, PythonExporter


def _read(wd, name):
    with open("{}/{}.ipynb".format(wd, name)) as f:
        notebook = nbformat.read(f, as_version=4)
    return notebook


def export_html(wd, name):
    nb = _read(wd, name)

    config = {
        'Exporter': {'template_file': 'embed',
                     'template_path': ['./sphinxext/']},
        'ExecutePreprocessor': {'enabled': True},
        'ExtractOutputPreprocessor': {'enabled': True},
        'CSSHTMLHeaderPreprocessor': {'enabled': True}
    }

    exporter = HTMLExporter(config)

    try:
        body, resources = exporter.from_notebook_node(nb)

        for fn, data in resources['outputs'].items():
            with open("{}/{}".format(wd, fn), 'wb') as f:
                f.write(data)
        return body
    except Exception as e:
        return str(e)


def export_python(wd, name):
    nb = _read(wd, name)
    exporter = PythonExporter()
    body, resources = exporter.from_notebook_node(nb)
    with open("{}/{}.py".format(wd, name), 'w') as f:
        f.write(body)


class NotebookDirective(Directive):
    """Insert an evaluated notebook into a document
    """
    required_arguments = 1
    optional_arguments = 1
    option_spec = {'skip_exceptions': directives.flag}
    final_argument_whitespace = True

    def run(self):

        # check if raw html is supported
        if not self.state.document.settings.raw_enabled:
            raise self.warning('"%s" directive disabled.' % self.name)

        # get path to notebook
        nb_rel_path = self.arguments[0]
        nb_abs_path = "{}/../{}".format(setup.confdir, nb_rel_path)
        nb_abs_path = os.path.abspath(nb_abs_path)
        nb_name = os.path.basename(nb_rel_path).split(".")[0]
        dest_dir = "{}/{}/{}".format(
            setup.app.builder.outdir,
            os.path.dirname(nb_rel_path),
            nb_name)
        fmt = {'wd': dest_dir, 'name': nb_name}

        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)

        shutil.copyfile(nb_abs_path, "{wd}/{name}.ipynb".format(**fmt))

        # TODO: Actually save evaluated notebook
        shutil.copyfile(nb_abs_path, "{wd}/{name}_eval.ipynb".format(**fmt))

        html = export_html(**fmt)
        export_python(**fmt)

        # Create link to notebook and script files
        link_rst = "({uneval}; {eval}; {py})".format(
            uneval=formatted_link("{wd}/{name}.ipynb".format(**fmt)),
            eval=formatted_link("{wd}/{name}_eval.ipynb".format(**fmt)),
            py=formatted_link("{wd}/{name}.py".format(**fmt)),
        )

        rst_file = self.state_machine.document.attributes['source']
        self.state_machine.insert_input([link_rst], rst_file)

        # create notebook node
        attributes = {'format': 'html', 'source': 'nb_path'}
        nb_node = notebook_node('', html, **attributes)
        nb_node.source, nb_node.line = self.state_machine \
            .get_source_and_line(self.lineno)

        # add dependency
        self.state.document.settings.record_dependencies.add(nb_abs_path)

        return [nb_node]


class notebook_node(nodes.raw):
    pass


def formatted_link(path):
    return "`%s <%s>`__" % (os.path.basename(path), path)


def visit_notebook_node(self, node):
    self.visit_raw(node)


def depart_notebook_node(self, node):
    self.depart_raw(node)


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    app.add_node(notebook_node,
                 html=(visit_notebook_node, depart_notebook_node))

    app.add_directive('notebook', NotebookDirective)



# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True

##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import sys

##############################################################################
# Headers
##############################################################################

cdef extern from "geometry.h" nogil:
    void dist(const float* xyz, const int* pairs, float* distance_out,
              float* displacement_out, int n_frames,  int n_atoms,
              int n_pairs)
    void dist_mic(const float* xyz, const int* pairs, const float* box_matrix,
                  float* distance_out, float* displacement_out,
                  int n_frames, int n_atoms, int n_pairs)
    void dist_t(const float* xyz, const int* pairs, const int* times,
                    float* distance_out, float* displacement_out, const int n_times, 
                    const int n_atoms, const int n_pairs)
    void dist_mic_t(const float* xyz, const int* pairs, const int* times,
                    const float* box_matrix, float* distance_out,
                    float* displacement_out, const int n_times, const int n_atoms,
                    const int n_pairs)
    void dist_mic_triclinic(const float* xyz, const int* pairs,
                            const float* box_matrix, float* distance_out,
                            float* displacement_out, int n_frames, int n_atoms,
                            int n_pairs)
    void dist_mic_triclinic_t(const float* xyz, const int* pairs, const int* times,
                              const float* box_matrix, float* distance_out,
                              float* displacement_out, int n_frames, int n_atoms,
                              int n_pairs)

    void angle(const float* xyz, const int* triplets, float* out,
               int n_frames, int n_atoms, int n_angles)

    void angle_mic(const float* xyz, const int* triplets, const float* box_matrix,
                   float* out, int n_frames, int n_atoms, int n_angles)

    void angle_mic_triclinic(const float* xyz, const int* triplets, const float* box_matrix,
                             float* out, int n_frames, int n_atoms, int n_angles)

    void dihedral(const float* xyz, const int* quartets, float* out,
                  int n_frames, int n_atoms,  int n_quartets)

    void dihedral_mic(const float* xyz, const int* quartets, float* out,
                      const float* box_matrix, int n_frames, int n_atoms,
                      int n_quartets)

    void dihedral_mic_triclinic(const float* xyz, const int* quartets, float* out,
                                const float* box_matrix, int n_frames, int n_atoms,
                                int n_quartets)

    void kabsch_sander(float* xyz, int* nco_indices, int* ca_indices,
                       const int* is_proline, int n_frames, int n_atoms,
                       int n_residues, int* hbonds, float* henergies)

    void dssp(const float* xyz, const int* nco_indices, const int* ca_indices,
              const int* is_proline, const int* chains_ids, const int n_frames,
              const int n_atoms, const int n_residues, char* secondary)

    void find_closest_contact(const float* positions, const int* group1, const int* group2,
                              int n_group1, int n_group2, const float* box_vectors_pointer,
                              int* atom1, int* atom2, float* distance)

cdef extern from "sasa.h":
    void sasa(const int n_frames, const int n_atoms, const float* xyzlist,
              const float* atom_radii, const int n_sphere_points,
              const int* atom_mapping, const int n_groups, float* out) nogil


##############################################################################
# Wrappers
##############################################################################

def _dist(float[:, :, ::1] xyz,
          int[:, ::1] pairs,
          float[:, ::1] out):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    dist(&xyz[0,0,0], &pairs[0,0], &out[0,0], NULL, n_frames, n_atoms, n_pairs)


def _dist_displacement(float[:, :, ::1] xyz,
                       int[:, ::1] pairs,
                       float[:, :, ::1] out):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    dist(&xyz[0,0,0], <int*> &pairs[0,0], NULL, &out[0,0,0], n_frames, n_atoms, n_pairs)


def _dist_mic(float[:, :, ::1] xyz,
              int[:, ::1] pairs,
              float[:, :, ::1] box_matrix,
              float[:, ::1] out,
              orthogonal):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    if orthogonal:
        dist_mic(&xyz[0,0,0], &pairs[0,0], &box_matrix[0,0,0], &out[0,0], NULL, n_frames, n_atoms, n_pairs)
    else:
        dist_mic_triclinic(&xyz[0,0,0], &pairs[0,0], &box_matrix[0,0,0], &out[0,0], NULL, n_frames, n_atoms, n_pairs)


def _dist_t(float[:, :, ::1] xyz,
        int[:, ::1] pairs,
        int[:, ::1] times,
        float[:, ::1] out):
    cdef int n_times = times.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    dist_t(&xyz[0,0,0], &pairs[0,0], &times[0,0], &out[0,0], NULL, n_times, n_atoms, n_pairs)


def _dist_mic_t(float[:, :, ::1] xyz,
              int[:, ::1] pairs,
              int[:, ::1] times,
              float[:, :, ::1] box_matrix,
              float[:, ::1] out,
              orthogonal):
    cdef int n_times = times.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    if orthogonal:
        dist_mic_t(&xyz[0,0,0], &pairs[0,0], &times[0,0], &box_matrix[0,0,0], &out[0,0], NULL, n_times, n_atoms, n_pairs)
    else:
        dist_mic_triclinic_t(&xyz[0,0,0], &pairs[0,0], &times[0,0], &box_matrix[0,0,0], &out[0,0], NULL, n_times, n_atoms, n_pairs)


def _dist_mic_displacement(float[:, :, ::1] xyz,
                           int[:, ::1] pairs,
                           float[:, :, ::1] box_matrix,
                           float[:, :, ::1] out,
                           orthogonal):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_pairs = pairs.shape[0]
    if orthogonal:
        dist_mic(&xyz[0,0,0], <int*> &pairs[0,0], &box_matrix[0,0,0], NULL, &out[0,0, 0], n_frames, n_atoms, n_pairs)
    else:
        dist_mic_triclinic(&xyz[0,0,0], <int*> &pairs[0,0], &box_matrix[0,0,0], NULL, &out[0,0, 0], n_frames, n_atoms, n_pairs)


def _angle(float[:, :, ::1] xyz,
           int[:, ::1] triplets,
           float[:, ::1] out):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_angles = triplets.shape[0]
    angle(&xyz[0,0,0], &triplets[0,0], &out[0,0], n_frames, n_atoms, n_angles)


def _angle_mic(float[:, :, ::1] xyz,
               int[:, ::1] triplets,
               float[:, :, ::1] box_matrix,
               float[:, ::1] out,
               orthogonal):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_angles = triplets.shape[0]
    if orthogonal:
        angle_mic(&xyz[0,0,0], &triplets[0,0], &box_matrix[0,0,0], &out[0,0], n_frames, n_atoms, n_angles)
    else:
        angle_mic_triclinic(&xyz[0,0,0], &triplets[0,0], &box_matrix[0,0,0], &out[0,0], n_frames, n_atoms, n_angles)


def _dihedral(float[:, :, ::1] xyz,
              int[:, ::1] quartets,
              float[:, ::1] out):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_quartets = quartets.shape[0]
    dihedral(&xyz[0,0,0], <int*> &quartets[0,0], &out[0,0], n_frames, n_atoms, n_quartets)


def _dihedral_mic(float[:, :, ::1] xyz,
                  int[:, ::1] quartets,
                  float[:, :, ::1] box_matrix,
                  float[:, ::1] out,
                  orthogonal):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_quartets = quartets.shape[0]
    if orthogonal:
        dihedral_mic(&xyz[0,0,0], <int*> &quartets[0,0], &box_matrix[0,0,0], &out[0,0], n_frames, n_atoms, n_quartets)
    else:
        dihedral_mic_triclinic(&xyz[0,0,0], <int*> &quartets[0,0], &box_matrix[0,0,0], &out[0,0], n_frames, n_atoms, n_quartets)


def _kabsch_sander(float[:, :, ::1] xyz,
                   int[:, ::1] nco_indices,
                   int[::1] ca_indices,
                   int[::1] is_proline,
                   int[:, :, ::1] hbonds,
                   float[:, :, ::1] henergies):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_residues = ca_indices.shape[0]
    kabsch_sander(&xyz[0,0,0], &nco_indices[0,0], &ca_indices[0],
                  &is_proline[0], n_frames, n_atoms, n_residues,
                  &hbonds[0,0,0], &henergies[0,0,0])


def _sasa(float[:, :, ::1] xyz,
          float[::1] atom_radii,
          int n_sphere_points,
          int[::1] atom_outmapping,
          float[:, ::1] out):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    sasa(n_frames, n_atoms, &xyz[0,0,0], &atom_radii[0], n_sphere_points,
         &atom_outmapping[0], out.shape[1], &out[0,0])


def _dssp(float[:, :, ::1] xyz,
          int[:, ::1] nco_indices,
          int[::1] ca_indices,
          int[::1] is_proline,
          int[::1] chain_ids):
    cdef int n_frames = xyz.shape[0]
    cdef int n_atoms = xyz.shape[1]
    cdef int n_residues = ca_indices.shape[0]
    cdef char[::1] secondary = bytearray(n_frames*n_residues)
    dssp(&xyz[0,0,0], &nco_indices[0,0], &ca_indices[0],
         &is_proline[0], &chain_ids[0], n_frames, n_atoms,
         n_residues, &secondary[0])

    PY2 = sys.version_info[0] == 2
    value = str(secondary.base) if PY2 else secondary.base.decode('ascii')
    return value


def _find_closest_contact(float[:, ::1] positions,
          int[::1] group1,
          int[::1] group2,
          box):
    cdef int atom1
    cdef int atom2
    cdef float distance
    cdef float[:, ::1] box_vectors
    cdef float* box_vectors_pointer
    if box is not None:
        box_vectors = np.asarray(box, order='c')
        box_vectors_pointer = &box_vectors[0,0]
    else:
        box_vectors_pointer = NULL
    find_closest_contact(&positions[0,0], &group1[0], &group2[0], len(group1), len(group2), box_vectors_pointer, &atom1, &atom2, &distance)
    return (atom1, atom2, distance)

include "image_molecules.pxi"



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: TJ Lane
# Contributors: Robert McGibbon and Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import mdtraj as md
import numpy as np
from mdtraj.testing import eq

def test_prmtop_with_overflow_mdcrd(get_fn):
    traj = md.load(get_fn('overflow.crd'), top=get_fn('tz2.truncoct.parm7'))
    # Check that the coordinates that 'overflow' were parsed correctly
    check = np.asarray([-0.131, -221.878, 2.631]) / 10.0 # To nanometers
    eq(check, traj.xyz[-1][-7])



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Jason Swails, Matthew Harrigan
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


from __future__ import print_function, division
import numpy as np
import re, os, tempfile
from mdtraj.formats.pdb import pdbstructure
from mdtraj.formats.pdb.pdbstructure import PdbStructure
from mdtraj.testing import eq
from mdtraj import load, load_pdb
from mdtraj import Topology
import pytest
import warnings

fd, temp = tempfile.mkstemp(suffix='.pdb')
os.close(fd)
def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.unlink(temp)

def test_pdbread(get_fn):
    pdb = get_fn('native.pdb')
    p = load(pdb)

def test_pdbread_with_input_top(get_fn):
    pdb = get_fn('native.pdb')
    p_1 = load(pdb)

    p_2 = load(pdb, top=pdb)

    eq(p_1.xyz, p_2.xyz)
    eq(p_1.topology, p_2.topology)


def test_pdbwrite(get_fn):
    pdb = get_fn('native.pdb')
    p = load(pdb)
    p.save(temp)

    r = load(temp)
    eq(p.xyz, r.xyz)


def test_load_multiframe(get_fn):
    with open(get_fn('multiframe.pdb')) as f:
        pdb = PdbStructure(f)
        assert eq(len(pdb.models), 2)
        assert eq(len(pdb.models[0].chains), 1)
        assert eq(len(pdb.models[0].chains[0].residues), 3)
        assert eq(len(list(pdb.models[0].iter_atoms())), 22)

        assert eq(len(pdb.models[1].chains), 1)
        assert eq(len(pdb.models[1].chains[0].residues), 3)
        assert eq(len(list(pdb.models[1].iter_atoms())), 22)


    t = load(get_fn('multiframe.pdb'))
    assert eq(t.n_frames, 2)
    assert eq(t.n_atoms, 22)
    assert eq(t.xyz[0], t.xyz[1])


def test_4ZUO(get_fn):
    t = load(get_fn('4ZUO.pdb'))
    eq(t.n_frames, 1)
    eq(t.n_atoms, 6200)

    # this is a random line from the file
    #ATOM   1609  O   GLY A 201     -25.423  13.774 -25.234  1.00  8.92           O

    atom = list(t.top.atoms)[1525]
    eq(atom.element.symbol, 'O')
    eq(atom.residue.name, 'GLY')
    eq(atom.index, 1525)
    eq(t.xyz[0, 1525], np.array([-25.423,  13.774, -25.234]) / 10)  # converting to NM

    # this is atom 1577 in the PDB
    #CONECT 1577 5518
    #ATOM   1577  O   HIS A 197     -18.521   9.724 -32.805  1.00  8.81           O
    #HETATM 5518  K     K A 402     -19.609  10.871 -35.067  1.00  9.11           K
    atom = list(t.top.atoms)[1493]
    eq(atom.name, 'O')
    eq(atom.residue.name, 'HIS')
    eq([(a1.index, a2.index) for a1, a2 in t.top.bonds if a1.index == 1493 or a2.index == 1493],
       [(1492, 1493), (1493, 5129)])

    # that first bond is from a conect record


def test_2EQQ_0(get_fn):
    # this is an nmr structure with 20 models
    t = load(get_fn('2EQQ.pdb'))
    assert eq(t.n_frames, 20)
    assert eq(t.n_atoms, 423)
    assert eq(len(list(t.top.residues)), 28)


def test_1vii_solvated_with_ligand(get_fn):
    traj = load(get_fn("1vii_sustiva_water.pdb"))
    eq(len(list(traj.top.bonds)), 5156)
    eq(len([bond for bond in traj.top.bonds if bond[0].residue.name == 'LIG']), 32)
    traj.save(temp)
    traj = load(temp)
    eq(len(list(traj.top.bonds)), 5156)
    eq(len([bond for bond in traj.top.bonds if bond[0].residue.name == 'LIG']), 32)

def test_write_large(get_fn):
    traj = load(get_fn('native.pdb'))
    traj.xyz.fill(123456789)
    with pytest.raises(ValueError):
        traj.save(temp)


def test_write_large_2(get_fn):
    traj = load(get_fn('native.pdb'))
    traj.xyz.fill(-123456789)
    with pytest.raises(ValueError):
        traj.save(temp)

def test_pdbstructure_0():
    pdb_lines = [
        "ATOM    188  N   CYS A  42      40.714  -5.292  12.123  1.00 11.29           N  ",
        "ATOM    189  CA  CYS A  42      39.736  -5.883  12.911  1.00 10.01           C  ",
        "ATOM    190  C   CYS A  42      40.339  -6.654  14.087  1.00 22.28           C  ",
        "ATOM    191  O   CYS A  42      41.181  -7.530  13.859  1.00 13.70           O  ",
        "ATOM    192  CB  CYS A  42      38.949  -6.825  12.002  1.00  9.67           C  ",
        "ATOM    193  SG  CYS A  42      37.557  -7.514  12.922  1.00 20.12           S  "
    ]

    res = pdbstructure.Residue("CYS", 42)
    for l in pdb_lines:
        res._add_atom(pdbstructure.Atom(l))
    for i, atom in enumerate(res):
        eq(pdb_lines[i], str(atom))


def test_pdbstructure_1():
    pdb_lines = [
         "ATOM    188  N   CYS A  42      40.714  -5.292  12.123  1.00 11.29           N",
         "ATOM    189  CA  CYS A  42      39.736  -5.883  12.911  1.00 10.01           C",
         "ATOM    190  C   CYS A  42      40.339  -6.654  14.087  1.00 22.28           C",
         "ATOM    191  O   CYS A  42      41.181  -7.530  13.859  1.00 13.70           O",
         "ATOM    192  CB  CYS A  42      38.949  -6.825  12.002  1.00  9.67           C",
         "ATOM    193  SG  CYS A  42      37.557  -7.514  12.922  1.00 20.12           S"
         ]
    positions = np.array([
        [ 40.714,  -5.292,  12.123],
        [ 39.736,  -5.883,  12.911],
        [ 40.339,  -6.654,  14.087],
        [ 41.181,  -7.53,   13.859],
        [ 38.949,  -6.825,  12.002],
        [ 37.557,  -7.514,  12.922]
    ])

    res = pdbstructure.Residue("CYS", 42)
    for l in pdb_lines:
        res._add_atom(pdbstructure.Atom(l))

    for i, c in enumerate(res.iter_positions()):
        eq(c, positions[i])

def test_pdbstructure_2():
    atom = pdbstructure.Atom("ATOM   2209  CB  TYR A 299       6.167  22.607  20.046  1.00  8.12           C")
    expected = np.array([6.167, 22.607, 20.046])
    for i, c in enumerate(atom.iter_coordinates()):
        eq(expected[i], c)

def test_pdbstructure_3():
    loc = pdbstructure.Atom.Location(' ', [1,2,3], 1.0, 20.0, "XXX")
    expected = [1, 2, 3]
    for i, c in enumerate(loc):
        eq(expected[i], c)

def test_pdb_from_url():
    # load pdb from URL
    t1 = load_pdb('http://www.rcsb.org/pdb/files/4ZUO.pdb.gz')
    t2 = load_pdb('http://www.rcsb.org/pdb/files/4ZUO.pdb')
    eq(t1.n_frames, 1)
    eq(t2.n_frames, 1)
    eq(t1.n_atoms, 6200)
    eq(t2.n_atoms, 6200)

def test_3nch_conect(get_fn):
    # This has conect entries that use all available digits, good failure case.
    t1 = load_pdb(get_fn('3nch.pdb.gz'))
    top, bonds = t1.top.to_dataframe()
    bonds = dict(((a, b), 1) for (a, b, _, _) in bonds)
    eq(bonds[19782, 19783], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19784], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19785], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19786], 1)  # Check that last SO4 molecule has right bonds


def test_3nch_serial_resSeq(get_fn):
    # If you use zero-based indexing, this PDB has quite large gaps in residue and atom numbering, so it's a good test case.  See #528
    # Gold standard values obtained via
    # cat 3nch.pdb |grep ATM|tail -n 5
    # HETATM19787  S   SO4 D 804      -4.788  -9.395  22.515  1.00121.87           S
    # HETATM19788  O1  SO4 D 804      -3.815  -9.511  21.425  1.00105.97           O
    # HETATM19789  O2  SO4 D 804      -5.989  -8.733  21.999  1.00116.13           O
    # HETATM19790  O3  SO4 D 804      -5.130 -10.726  23.043  1.00108.74           O
    # HETATM19791  O4  SO4 D 804      -4.210  -8.560  23.575  1.00112.54           O
    t1 = load_pdb(get_fn('3nch.pdb.gz'))
    top, bonds = t1.top.to_dataframe()

    top2 = Topology.from_dataframe(top, bonds)
    eq(t1.top, top2)

    top = top.set_index('serial')  # Index by the actual data in the PDB
    eq(str(top.loc[19791]["name"]), "O4")
    eq(str(top.loc[19787]["name"]), "S")
    eq(str(top.loc[19787]["resName"]), "SO4")
    eq(int(top.loc[19787]["resSeq"]), 804)

def test_1ncw(get_fn):
    t1 = load_pdb(get_fn('1ncw.pdb.gz'))

def test_1vii_url_and_gz(get_fn):
    t1 = load_pdb('http://www.rcsb.org/pdb/files/1vii.pdb.gz')
    t2 = load_pdb('http://www.rcsb.org/pdb/files/1vii.pdb')
    t3 = load_pdb(get_fn('1vii.pdb.gz'))
    t4 = load_pdb(get_fn('1vii.pdb'))
    eq(t1.n_frames, 1)
    eq(t1.n_frames, t2.n_frames)
    eq(t1.n_frames, t3.n_frames)
    eq(t1.n_frames, t4.n_frames)

    eq(t1.n_atoms, t2.n_atoms)
    eq(t1.n_atoms, t3.n_atoms)
    eq(t1.n_atoms, t4.n_atoms)

def test_segment_id(get_fn):
    pdb = load_pdb(get_fn('ala_ala_ala.pdb'))
    pdb.save_pdb(temp)
    pdb2 = load_pdb(temp)

    correct_segment_id = 'AAL'
    # check that all segment ids are set correctly
    for ridx,r in enumerate(pdb.top.residues):
        assert r.segment_id == correct_segment_id, "residue %i (0-indexed) does not have segment_id set correctly from ala_ala_ala.pdb"%(ridx)

    # check that all segment ids are set correctly after a new pdb file is written
    for ridx,(r1,r2) in enumerate(zip(pdb.top.residues,pdb2.top.residues)):
        assert r1.segment_id == r2.segment_id, "segment_id of residue %i (0-indexed) in ala_ala_ala.pdb does not agree with value in after being written out to a new pdb file"%(ridx)


def test_bfactors(get_fn):
    pdb = load_pdb(get_fn('native.pdb'))
    bfactors0 = np.arange(pdb.n_atoms) / 2.0 - 4.0 # (Get some decimals..)

    pdb.save_pdb(temp, bfactors=bfactors0)

    with open(temp, 'r') as fh:
        atom_lines = [line for line in fh.readlines() if re.search(r'^ATOM', line)]

    str_bfactors1 = [l[60:66] for l in atom_lines]
    flt_bfactors1 = np.array([float(i) for i in str_bfactors1])

    # check formatting has a space at the beginning and not at the end
    frmt = np.array([(s[0] == ' ') and (s[-1] != ' ') for s in str_bfactors1])
    assert np.all(frmt)

    # make sure the numbers are actually the same
    eq(bfactors0, flt_bfactors1)

def test_hex(get_fn):
   pdb = load_pdb(get_fn('water_hex.pdb.gz'))
   assert pdb.n_atoms == 100569
   assert pdb.n_residues == 33523
   pdb.save(temp)


def test_dummy_pdb_box_detection(get_fn, recwarn):
    traj = load(get_fn('2koc.pdb'))
    assert len(recwarn) == 1
    w = recwarn.pop(UserWarning)
    assert 'Unlikely unit cell' in str(w.message)
    assert traj.unitcell_lengths is None, 'Expected dummy box to be deleted'


def test_multichain_load_cycle(get_fn):
    # Issue 1611, make sure that save/load works for more than 1 chain
    pdb = load(get_fn('issue_1611.pdb'))
    bonds = [(bond.atom1.index, bond.atom2.index)
             for bond in pdb.topology.bonds]
    pdb.save(temp)
    pdb2 = load_pdb(temp)
    bonds2 = [(bond.atom1.index, bond.atom2.index)
              for bond in pdb2.topology.bonds]
    assert len(bonds) == len(bonds2)

def test_load_pdb_input_top(get_fn):

    pdb = get_fn('native.pdb')
    p_1 = load_pdb(pdb)

    p_2 = load_pdb(pdb, top=p_1.topology)

    eq(p_1.xyz, p_2.xyz)
    eq(p_1.topology, p_2.topology)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import numpy as np
from mdtraj.geometry import alignment
from mdtraj.testing import eq

xyz1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0], [1, 0, 1], [0, 1, 1.0]])

offset = 1.0 * np.random.normal(size=(3))
rotation = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
xyz2 = rotation.dot(xyz1.T).T + offset
xyz3 = xyz1 + np.random.normal(size=xyz1.shape)
extra_points = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15]])
xyz4 = rotation.dot(np.concatenate([xyz1, extra_points]).T).T + offset


def test_rmsd_zero():
    rmsd_kabsch = alignment.rmsd_kabsch(xyz1, xyz2)
    rmsd_qcp = alignment.rmsd_qcp(xyz1, xyz2)
    eq(float(rmsd_kabsch), 0.0, decimal=5)
    eq(float(rmsd_qcp), 0.0, decimal=5)


def test_rmsd_nonzero():
    rmsd_kabsch = alignment.rmsd_kabsch(xyz1, xyz3)
    rmsd_qcp = alignment.rmsd_qcp(xyz1, xyz3)
    eq(rmsd_kabsch, rmsd_qcp, decimal=5)


def test_transform():
    T = alignment.compute_transformation(xyz2, xyz1)
    xyz2_prime = T.transform(xyz2)

    eq(xyz1, xyz2_prime)


def test_transform2():
    # For testing transform on all points when only subset for align
    xyz1_len = np.shape(xyz1)[0]
    T = alignment.compute_transformation(xyz4[0:xyz1_len, :], xyz1)
    xyz4_prime = T.transform(xyz4)

    eq(xyz1, xyz4_prime[0:xyz1_len, :])



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Jason Swails
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import os
import subprocess
from distutils.spawn import find_executable

import mdtraj as md
import pytest
from mdtraj.formats import psf
from mdtraj.testing import eq
from mdtraj.utils import enter_temp_directory

VMD = find_executable('vmd')
needs_vmd = pytest.mark.skipif(
    not find_executable('vmd'),
    reason='This test requires the VMD executable: http://www.ks.uiuc.edu/Research/vmd/'
)


def test_load_psf(get_fn):
    top = psf.load_psf(get_fn('ala_ala_ala.psf'))
    ref_top = md.load(get_fn('ala_ala_ala.pdb')).topology
    eq(top, ref_top)
    # Test the XPLOR psf format parsing
    top2 = psf.load_psf(get_fn('ala_ala_ala.xpsf'))
    eq(top2, ref_top)
    # Test segment_names are loaded properly
    assert next(top.residues).segment_id == "AAL", "Segment id is not being assigned correctly for ala_ala_ala.psf"
    assert next(top2.residues).segment_id == "AAL", "Segment id is not being assigned correctly for ala_ala_ala.xpsf"


def test_multichain_psf(get_fn):
    top = psf.load_psf(get_fn('3pqr_memb.psf'))
    # Check that each segment was turned into a chain
    eq(top.n_chains, 11)
    chain_lengths = [5162, 185, 97, 28, 24, 24, 45, 35742, 72822, 75, 73]
    for i, n in enumerate(chain_lengths):
        eq(top.chain(i).n_atoms, n)
    # There are some non-sequentially numbered residues across chain
    # boundaries... make sure resSeq is properly taken from the PSF
    eq(top.chain(0).residue(0).resSeq, 1)
    eq(top.chain(0).residue(-1).resSeq, 326)
    eq(top.chain(1).residue(0).resSeq, 340)
    eq(top.chain(1).residue(-1).resSeq, 350)
    eq(top.chain(2).residue(0).resSeq, 1)
    eq(top.chain(2).residue(-1).resSeq, 4)
    eq(top.chain(3).residue(0).resSeq, 1)
    eq(top.chain(3).residue(-1).resSeq, 1)
    eq(top.chain(4).residue(0).resSeq, 1)
    eq(top.chain(4).residue(-1).resSeq, 1)
    eq(top.chain(5).residue(0).resSeq, 1)
    eq(top.chain(5).residue(-1).resSeq, 1)
    eq(top.chain(6).residue(0).resSeq, 1)
    eq(top.chain(6).residue(-1).resSeq, 2)
    eq(top.chain(7).residue(0).resSeq, 1)
    eq(top.chain(7).residue(-1).resSeq, 276)
    eq(top.chain(8).residue(0).resSeq, 1)
    eq(top.chain(8).residue(-1).resSeq, 24274)
    eq(top.chain(9).residue(0).resSeq, 1)
    eq(top.chain(9).residue(-1).resSeq, 75)
    eq(top.chain(10).residue(0).resSeq, 1)
    eq(top.chain(10).residue(-1).resSeq, 73)


def test_load_mdcrd_with_psf(get_fn):
    traj = md.load(get_fn('ala_ala_ala.mdcrd'), top=get_fn('ala_ala_ala.psf'))
    ref_traj = md.load(get_fn('ala_ala_ala.pdb'))
    eq(traj.topology, ref_traj.topology)
    eq(traj.xyz, ref_traj.xyz)


@needs_vmd
@pytest.mark.parametrize('pdb', ['1vii.pdb', '2EQQ.pdb'])
def test_against_vmd(pdb, get_fn):
    pdb = get_fn(pdb)
    # this is probably not cross-platform compatible. I assume that the exact
    # path to this CHARMM topology that is included with VMD depends on
    # the install mechanism, especially for bundled mac or windows installers
    VMD_ROOT = os.path.join(os.path.dirname(os.path.realpath(VMD)), '..')
    top_paths = [os.path.join(r, f) for (r, _, fs) in os.walk(VMD_ROOT) for f in fs
                 if 'top_all27_prot_lipid_na.inp' in f]
    assert len(top_paths) >= 0
    top = os.path.abspath(top_paths[0]).replace(" ", "\\ ")

    TEMPLATE = '''
package require psfgen
topology %(top)s
pdbalias residue HIS HSE
pdbalias atom ILE CD1 CD
segment U {pdb %(pdb)s}
coordpdb %(pdb)s U
guesscoord
writepdb out.pdb
writepsf out.psf
exit
    ''' % {'top': top, 'pdb': pdb}

    with enter_temp_directory():
        with open('script.tcl', 'w') as f:
            f.write(TEMPLATE)
        subprocess.check_call([VMD, '-startup', 'script.tcl', '-dispdev', 'none'])
        out_pdb = md.load('out.pdb')
        out_psf = md.load_psf('out.psf')

        # make sure the two topologies are equal
        eq(out_pdb.top, out_psf)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import os
import subprocess
from distutils.spawn import find_executable

import mdtraj as md
import numpy as np
import pytest
import scipy.sparse
from mdtraj.testing import eq

DSSP_MSG = "This test requires mkdssp to be installed, from http://swift.cmbi.ru.nl/gv/dssp/"
needs_dssp = pytest.mark.skipif(not find_executable('mkdssp'), reason=DSSP_MSG)


def test_hbonds(get_fn):
    t = md.load(get_fn('2EQQ.pdb'))
    ours = md.geometry.hbond.kabsch_sander(t)


@needs_dssp
def test_hbonds_against_dssp(get_fn, tmpdir):
    t = md.load(get_fn('2EQQ.pdb'))[0]
    pdb = os.path.join(tmpdir, 'f.pdb')
    dssp = os.path.join(tmpdir, 'f.pdb.dssp')
    t.save(pdb)

    cmd = ['mkdssp', '-i', pdb, '-o', dssp]
    subprocess.check_output(' '.join(cmd), shell=True)
    energy = scipy.sparse.lil_matrix((t.n_residues, t.n_residues))

    # read the dssp N-H-->O column from the output file
    with open(dssp) as f:
        # skip the lines until the description of each residue's hbonds
        while not f.readline().startswith('  #  RESIDUE AA STRUCTURE'):
            continue

        for i, line in enumerate(f):
            line = line.rstrip()
            offset0, e0 = map(float, line[39:50].split(','))
            offset1, e1 = map(float, line[61:72].split(','))
            if e0 <= -0.5:
                energy[int(i + offset0), i] = e0
            if e1 <= -0.5:
                energy[int(i + offset1), i] = e1

    dssp = energy.todense()
    ours = md.geometry.hbond.kabsch_sander(t)[0].todense()

    # There is tricky issues with the rounding right at the -0.5 cutoff,
    # so lets just check for equality with DSSP at -0.6 or less
    eq((dssp < -0.6), (ours < -0.6))
    eq(dssp[dssp < -0.6], ours[ours < -0.6], decimal=1)


def test_baker_hubbard_0(get_fn):
    t = md.load(get_fn('2EQQ.pdb'))

    # print('to view the hbonds defined in 2EQQ by baker_hubbard()')
    # print('put these commands into pymol on top of the pdb:\n')
    # for e in md.geometry.hbond.baker_hubbard(t):
    #     print('distance RANK %d, RANK %d' % (e[1], e[2]))

    # these are the results produced by the algorithm on this protein as
    # of 11/26/13. This unit test basically just ensures that the method
    # runs and produces the same results it did then. It's no guarentee that
    # these are the "TRUE" hydrogen bonds in this system.
    ref = np.array([[0, 10, 8], [0, 11, 7], [69, 73, 54], [76, 82, 65],
                    [119, 131, 89], [140, 148, 265], [166, 177, 122],
                    [181, 188, 231], [209, 217, 215], [221, 225, 184],
                    [228, 239, 186], [235, 247, 216], [262, 271, 143],
                    [298, 305, 115], [186, 191, 215], [413, 419, 392]])
    eq(ref, md.geometry.hbond.baker_hubbard(t))


def test_baker_hubbard_1(get_fn):
    # no hydrogens in this file -> no hydrogen bonds
    t = md.load(get_fn('1bpi.pdb'))
    eq(np.zeros((0, 3), dtype=int), md.baker_hubbard(t))


def test_baker_hubbard_2(get_fn):
    t = md.load(get_fn('1vii_sustiva_water.pdb'))
    triplets = md.baker_hubbard(t)
    N = 1000
    rows = triplets[:, 0] * N * N + triplets[:, 1] * N + triplets[:, 2]
    # ensure that there aren't any repeat rows
    eq(len(np.unique(rows)), len(rows))


def test_baker_hubbard_3(get_fn):
    #different distance cutoffs->different hydrogen bonds found
    t = md.load(get_fn('2waters_baker_hubbard.pdb'))
    eq(np.zeros((0,3), dtype=int), md.baker_hubbard(t, exclude_water=False, distance_cutoff = 0.18))
    eq(np.zeros((0,3), dtype=int), md.baker_hubbard(t, exclude_water=False, angle_cutoff=180))
    eq(np.array([[0, 1, 3]]), md.baker_hubbard(t, exclude_water=False))


def test_wernet_nilsson_0(get_fn):
    # no hydrogens in this file -> no hydrogen bonds
    t0 = md.load(get_fn('1bpi.pdb'))
    assert len(md.wernet_nilsson(t0)) == len(t0)
    eq(np.zeros((0, 3), dtype=int), md.wernet_nilsson(t0)[0])


def test_wernet_nilsson_1(get_fn):
    # one of these files has PBCs and the other doesnt
    for fn in ['2EQQ.pdb', '4ZUO.pdb']:
        t = md.load(get_fn(fn))

        result = md.wernet_nilsson(t)
        assert len(result) == len(t)
        assert isinstance(result, list)
        assert all(isinstance(elem, np.ndarray) for elem in result)
        assert all(elem.shape[1] == 3 for elem in result)
        for frame, hbonds in enumerate(result):
            for d_i, h_i, a_i in hbonds:
                assert t.topology.atom(d_i).element.symbol in ['O', 'N']
                assert t.topology.atom(h_i).element.symbol == 'H'
                assert t.topology.atom(a_i).element.symbol in ['O', 'N']
            # assert that the donor-acceptor distance is less than 0.5 nm, just
            # to make sure the criterion is giving back totally implausible stuff
            if len(hbonds) > 0:
                assert np.all(md.compute_distances(t[frame], hbonds[:, [0, 2]]) < 0.5)



import mdtraj as md
import numpy as np

random = np.random.RandomState(0)


def _run_one_test(periodic, triclinic):
    n_atoms = 100
    cutoff = 1.5
    box_size = np.array([3.0, 4.0, 5.0])
    xyz = random.rand(1, n_atoms, 3) * box_size
    if periodic:
        unitcell_lengths = box_size
    else:
        unitcell_lengths = None
    if triclinic:
        unitcell_angles = np.array([80.0, 90.0, 100.0])
    else:
        unitcell_angles = None
    traj = md.Trajectory(xyz=xyz, topology=None, unitcell_lengths=unitcell_lengths, unitcell_angles=unitcell_angles)
    neighbors = md.compute_neighborlist(traj, cutoff)
    pairs = np.array([(i, j) for i in range(n_atoms) for j in range(n_atoms) if i < j])
    dists = md.compute_distances(traj, pairs, True, False)[0]
    for (i, j), d in zip(pairs, dists):
        if d < cutoff:
            assert (j in neighbors[i])
            assert (i in neighbors[j])
        else:
            assert (j not in neighbors[i])
            assert (i not in neighbors[j])


def test_nonperiodic():
    _run_one_test(False, False)


def test_periodic():
    _run_one_test(True, False)


def test_triclinic():
    _run_one_test(True, True)
